/*
*******************************************************************
* Header file for B002_IDE.C - IDE interface for Compact Flash, Hardisk, etc.
* for TS02 II
* File:  B002_IDE.H
* Autor: DevCom sro, Na Jarove 50, Praha 3, Jan Damborsky
* Creation Date: 24.5.2004
* Last modification: 28.7.2004
* Version: v1.0
* List of changes:
*******************************************************************
*/

#ifndef _B002_IDE_HEADER_
#define _B002_IDE_HEADER_

#include "target.h"

/* DEFINICE KONSTANT */
/* return codes */
#define IDE_RETOK		0x00	/* v poradku */
#define IDE_ERRIDENT	-1		/* chyba pri cteni identifikace */

/* IO piny  */
#define IDE_RESET_PIN		0x80	/* PC[7] */

#define CFG_IDE_MAXBUS		1	/* max. 1 IDE bus		*/
#define CFG_IDE_MAXDEVICE	1	/* max. 1 drive per IDE bus	*/

#define CFG_ATA_IDE0_OFFSET	0x0000

#define CFG_ATA_BASE_ADDR	IDE_BASE_ADDRESS
#define CFG_ATA_ALT_OFFSET	IDE_ALTSTS_BASE_ADDRESS
#define CFG_ATA_DATA_OFFSET	0
#define CFG_ATA_REG_OFFSET	0

// baze IDE registru
#define ATA_IO_DATA(x)	(CFG_ATA_DATA_OFFSET+(x))
#define ATA_IO_REG(x)	(CFG_ATA_REG_OFFSET + 2*(x))
#define ATA_IO_ALT(x)	(CFG_ATA_ALT_OFFSET + 2*(x))

// adresy registru
#define ATA_DATA_REG	ATA_IO_DATA(0)
#define ATA_ERROR_REG	ATA_IO_REG(1)
#define ATA_SECT_CNT	ATA_IO_REG(2)
#define ATA_SECT_NUM	ATA_IO_REG(3)
#define ATA_CYL_LOW		ATA_IO_REG(4)
#define ATA_CYL_HIGH	ATA_IO_REG(5)
#define ATA_DEV_HD		ATA_IO_REG(6)
#define ATA_COMMAND		ATA_IO_REG(7)
#define ATA_STATUS		ATA_COMMAND
#define ATA_DEV_CTL		ATA_IO_ALT(6)
#define ATA_LBA_LOW		ATA_SECT_NUM
#define ATA_LBA_MID		ATA_CYL_LOW
#define ATA_LBA_HIGH	ATA_CYL_HIGH
#define ATA_LBA_SEL		ATA_DEV_CTL

/*
 * Status register bits
 */
#define ATA_STAT_BUSY	0x80	/* Device Busy			*/
#define ATA_STAT_READY	0x40	/* Device Ready			*/
#define ATA_STAT_FAULT	0x20	/* Device Fault			*/
#define ATA_STAT_SEEK	0x10	/* Device Seek Complete		*/
#define ATA_STAT_DRQ	0x08	/* Data Request (ready)		*/
#define ATA_STAT_CORR	0x04	/* Corrected Data Error		*/
#define ATA_STAT_INDEX	0x02	/* Vendor specific		*/
#define ATA_STAT_ERR	0x01	/* Error			*/

/*
 * Device / Head Register Bits
 */
#define ATA_DEVICE(x)	((x & 1)<<4)
#define ATA_LBA		0xE0

/*
 * ATA Commands (only mandatory commands listed here)
 */
#define ATA_CMD_READ	0x20	/* Read Sectors (with retries)	*/
#define ATA_CMD_READN	0x21	/* Read Sectors ( no  retries)	*/
#define ATA_CMD_WRITE	0x30	/* Write Sectores (with retries)*/
#define ATA_CMD_WRITEN	0x31	/* Write Sectors  ( no  retries)*/
#define ATA_CMD_VRFY	0x40	/* Read Verify  (with retries)	*/
#define ATA_CMD_VRFYN	0x41	/* Read verify  ( no  retries)	*/
#define ATA_CMD_SEEK	0x70	/* Seek				*/
#define ATA_CMD_DIAG	0x90	/* Execute Device Diagnostic	*/
#define ATA_CMD_INIT	0x91	/* Initialize Device Parameters	*/
#define ATA_CMD_RD_MULT	0xC4	/* Read Multiple		*/
#define ATA_CMD_WR_MULT	0xC5	/* Write Multiple		*/
#define ATA_CMD_SETMULT	0xC6	/* Set Multiple Mode		*/
#define ATA_CMD_RD_DMA	0xC8	/* Read DMA (with retries)	*/
#define ATA_CMD_RD_DMAN	0xC9	/* Read DMS ( no  retries)	*/
#define ATA_CMD_WR_DMA	0xCA	/* Write DMA (with retries)	*/
#define ATA_CMD_WR_DMAN	0xCB	/* Write DMA ( no  retires)	*/
#define ATA_CMD_IDENT	0xEC	/* Identify Device		*/
#define ATA_CMD_SETF	0xEF	/* Set Features			*/
#define ATA_CMD_CHK_PWR	0xE5	/* Check Power Mode		*/

#define IF_TYPE_IDE			1
#define IF_TYPE_SCSI		2
#define IF_TYPE_ATAPI		3
#define IF_TYPE_USB			4
#define IF_TYPE_DOC			5

/* PIO */
#define	IDE_MAX_PIO_MODE	4	/* max suppurted PIO mode		*/

/* default PIO mod */
#ifndef	CFG_PIO_MODE
#define	CFG_PIO_MODE	0		/* use a relaxed default */
#endif

#define ATA_BLOCKSIZE	512	/* bytes */
#define ATA_BLOCKSHIFT	9	/* 2 ^ ATA_BLOCKSIZESHIFT = 512 */
#define ATA_SECTORWORDS	(512 / sizeof(unsigned long))

#define IDE_TIME_OUT	2000	/* 2 sec timeout */
#define ATAPI_TIME_OUT	7000	/* 7 sec timeout (5 sec seems to work...) */
#define IDE_SPIN_UP_TIME_OUT 5000 /* 5 sec spin-up timeout */

#define DEV_TYPE_UNKNOWN 0xff /* not connected */
#define DEV_TYPE_HARDDISK 0x00 /* harddisk */
#define DEV_TYPE_TAPE 		0x01 /* Tape */
#define DEV_TYPE_CDROM 		0x05 /* CD-ROM */
#define DEV_TYPE_OPDISK 	0x07 /* optical disk */

#define	PART_TYPE_UNKNOWN	0x00
#define PART_TYPE_MAC		0x01
#define PART_TYPE_DOS		0x02
#define PART_TYPE_ISO		0x03

/* DEFINICE TYPU */
/* deskriptor pro popis IDE zarizeni */
typedef struct block_dev_desc {
	int           if_type;    /* type of the interface */
	int           dev;        /* device number */
	unsigned char part_type;  /* partition type */
	unsigned char target;			/* target SCSI ID */
	unsigned char lun;				/* target LUN */
	unsigned char type;				/* device type */
	unsigned long lba;        /* number of blocks */
	unsigned long blksz;			/* block size */
	unsigned char vendor[40]; /* IDE model, SCSI Vendor */
	unsigned char product[20];/* IDE Serial no, SCSI product */
	unsigned char revision[8];/* firmware revision */
	unsigned char removable;	/* removable device */
	unsigned long (*block_read)(int dev,unsigned long start,unsigned long blkcnt, unsigned long *buffer);
} block_dev_desc_t;

/* struktura pro popis PIO casovani */
typedef struct {
	unsigned int	t_setup;	/* Setup  Time in [ns] or clocks	*/
	unsigned int	t_length;	/* Length Time in [ns] or clocks	*/
	unsigned int	t_hold;		/* Hold   Time in [ns] or clocks	*/
} pio_config_t;

/* struktura identifikacniho zaznamu nacteneneho prikazem IDENT (0xec) */
typedef struct hd_driveid {
	unsigned short	config;		/* lots of obsolete bit flags */
	unsigned short	cyls;		/* "physical" cyls */
	unsigned short	reserved2;	/* reserved (word 2) */
	unsigned short	heads;		/* "physical" heads */
	unsigned short	track_bytes;	/* unformatted bytes per track */
	unsigned short	sector_bytes;	/* unformatted bytes per sector */
	unsigned short	sectors;	/* "physical" sectors per track */
	unsigned short	vendor0;	/* vendor unique */
	unsigned short	vendor1;	/* vendor unique */
	unsigned short	vendor2;	/* vendor unique */
	unsigned char	serial_no[20];	/* 0 = not_specified */
	unsigned short	buf_type;
	unsigned short	buf_size;	/* 512 byte increments; 0 = not_specified */
	unsigned short	ecc_bytes;	/* for r/w long cmds; 0 = not_specified */
	unsigned char	fw_rev[8];	/* 0 = not_specified */
	unsigned char	model[40];	/* 0 = not_specified */
	unsigned char	max_multsect;	/* 0=not_implemented */
	unsigned char	vendor3;	/* vendor unique */
	unsigned short	dword_io;	/* 0=not_implemented; 1=implemented */
	unsigned char	vendor4;	/* vendor unique */
	unsigned char	capability;	/* bits 0:DMA 1:LBA 2:IORDYsw 3:IORDYsup*/
	unsigned short	reserved50;	/* reserved (word 50) */
	unsigned char	vendor5;	/* vendor unique */
	unsigned char	tPIO;		/* 0=slow, 1=medium, 2=fast */
	unsigned char	vendor6;	/* vendor unique */
	unsigned char	tDMA;		/* 0=slow, 1=medium, 2=fast */
	unsigned short	field_valid;	/* bits 0:cur_ok 1:eide_ok */
	unsigned short	cur_cyls;	/* logical cylinders */
	unsigned short	cur_heads;	/* logical heads */
	unsigned short	cur_sectors;	/* logical sectors per track */
	unsigned short	cur_capacity0;	/* logical total sectors on drive */
	unsigned short	cur_capacity1;	/*  (2 words, misaligned int)     */
	unsigned char	multsect;	/* current multiple sector count */
	unsigned char	multsect_valid;	/* when (bit0==1) multsect is ok */
	unsigned int	lba_capacity;	/* total number of sectors */
	unsigned short	dma_1word;	/* single-word dma info */
	unsigned short	dma_mword;	/* multiple-word dma info */
	unsigned short  eide_pio_modes; /* bits 0:mode3 1:mode4 */
	unsigned short  eide_dma_min;	/* min mword dma cycle time (ns) */
	unsigned short  eide_dma_time;	/* recommended mword dma cycle time (ns) */
	unsigned short  eide_pio;       /* min cycle time (ns), no IORDY  */
	unsigned short  eide_pio_iordy; /* min cycle time (ns), with IORDY */
	unsigned short	words69_70[2];	/* reserved words 69-70 */
	unsigned short	words71_74[4];	/* reserved words 71-74 */
	unsigned short  queue_depth;	/*  */
	unsigned short  words76_79[4];	/* reserved words 76-79 */
	unsigned short  major_rev_num;	/*  */
	unsigned short  minor_rev_num;	/*  */
	unsigned short  command_set_1;	/* bits 0:Smart 1:Security 2:Removable 3:PM */
	unsigned short  command_set_2;	/* bits 14:Smart Enabled 13:0 zero */
	unsigned short  cfsse;		/* command set-feature supported extensions */
	unsigned short  cfs_enable_1;	/* command set-feature enabled */
	unsigned short  cfs_enable_2;	/* command set-feature enabled */
	unsigned short  csf_default;	/* command set-feature default */
	unsigned short  dma_ultra;	/*  */
	unsigned short	word89;		/* reserved (word 89) */
	unsigned short	word90;		/* reserved (word 90) */
	unsigned short	CurAPMvalues;	/* current APM values */
	unsigned short	word92;		/* reserved (word 92) */
	unsigned short	hw_config;	/* hardware config */
	unsigned short  words94_125[32];/* reserved words 94-125 */
	unsigned short	last_lun;	/* reserved (word 126) */
	unsigned short	word127;	/* reserved (word 127) */
	unsigned short	dlf;		/* device lock function
					 * 15:9	reserved
					 * 8	security level 1:max 0:high
					 * 7:6	reserved
					 * 5	enhanced erase
					 * 4	expire
					 * 3	frozen
					 * 2	locked
					 * 1	en/disabled
					 * 0	capability
					 */
	unsigned short  csfo;		/* current set features options
					 * 15:4	reserved
					 * 3	auto reassign
					 * 2	reverting
					 * 1	read-look-ahead
					 * 0	write cache
					 */
	unsigned short	words130_155[26];/* reserved vendor words 130-155 */
	unsigned short	word156;
	unsigned short	words157_159[3];/* reserved vendor words 157-159 */
	unsigned short	words160_255[95];/* reserved words 160-255 */
} hd_driveid_t;

/* DEFINICE MAKER */
// prevod cisla zarizeni na cislo sbernice - na kazde sbernici max. 2 zarizeni
#define	IDE_BUS(dev)	(dev >> 1)

/* vypocet poctu cyklu pro nastaveni PIO modu */
#define PCMCIA_MK_CLKS( t, T ) (( (t) * ((T)/1000000U) + 999U ) / 1000U)

#ifndef ATA_RESET_TIME
#define ATA_RESET_TIME	60	/* spec allows up to 31 seconds */
#endif

/*****/
// vraci 0, v pripade neuspechu vraci -1
#define ReadSector(drive, startinglogicalsector, numberofsectors, buffer) \
 (ide_read((int)(drive), (ulong)(startinglogicalsector), (ulong)(numberofsectors), \
 (ulong *)(buffer)) == (ulong)(numberofsectors) ? 0 : -1)

/* DEKLARACE EXTERNICH PROMENNYCH */
// deskriptor IDE zarizeni
extern block_dev_desc_t ide_dev_desc[CFG_IDE_MAXDEVICE];

/* DEKLARACE HLAVICEK FUNKCI */
/*****/
// inicializace IDE rozhrani
// vraci: IDE_RETOK...v poradku
int
ide_init(void);

/*****/
// nacte identifikaci IDE zarizeni
// vraci: IDE_RETOK..identifikace nactena v poradku
//        IDE_ERRIDENT..chyba pri cteni identifikace
int
ide_read_ident(
	int		device,		// cislo zarizeni
	ulong	*buffer		// ukazatel na buffer pro nactena data
);

/*****/
// nacte sektory z IDE zarizeni
// vraci: pocet nactenych bloku
ulong
ide_read(
	int		device,		// cislo IDE zarizeni
	ulong	blknr,		// cislo prvniho bloku pro cteni
	ulong	blkcnt,		// pocet bloku pro nacteni
	ulong	*buffer		// ukazatel na buffer pro nactena data
);

/*****/
// zapise sektory na IDE zarizeni
// vraci: pocet zapsanych sektoru
ulong
ide_write(
	int		device,		// cislo zarizeni
	ulong	blknr,		// cislo prvniho bloku pro zapis
	ulong	blkcnt,		// pocet bloku pro zapis
	ulong	*buffer		// ukazatel na buffer
);

#endif /* _B002_IDE_HEADER_ */
