/****************************************************************/
/*                   Downloader do FLASH                        */
/*                    Power MPC555                              */  
/*  Lubos Jelnek                                               */  
/*  posl. zmeny: 5.8.2004 by !HA!                               */  
/*                                                              */  
/*  Nahrv do FLASH, lze na stavit ku slova 1-4 byte     */  
/*  Pouziva SMC1                                                */  
/*  Sedi v RAM a nahrava do interni a externi FLASH             */
/*  Spousti aplikaci v interni FLASH na druhem sektoru = 0x8000 */  
/*  nedela inicializaci UARTU natvrdo!!                         */            
/****************************************************************/
//*
//*  Verze pro FLASH CDR MoneT + AM29F040      
//*   Pouze DownLoad, a rts!!
//*
//*   Nahrava po bytech soubor Motorola Bin
//*
//*   Nejdrive kontroluje Serial number !!
//*
//*   Dela Chip erase, nebo sector erase
//*   Dela Echo kazdeho znaku v Mot file, aby brzdil prenos!!
//*
//*   Nejprve ceka na $12, vysilany periodicky z PC
//*    pak posle 'xxxx' a ceka na $21, ktery echuje
//*    a prvni radek s Serial_number
//*   Kdyz OK,posle $22, maze a posila $23
//*   Pak Mot file
//*   Radek = Delka,Adresa,Data,Suma  //Binary
//*   Pokud Adresa=-1 konec zaznamu
//*
//****************************************************************
//****************************************************************

#include "eth.h"
#include "ip.h"
#include "udp.h"

/* velikost ser. cisla */
#define SERIAL_COUNT 14

/* loader v1.2 - rychlejsi - echuje pouze sumu, provadi Xorovani bytu */
#define FAST_LOADER		1
//#undef FAST_LOADER		

/* timeout v [ms] pri cekani na prijem znaku */
#define cLOADER_RECV_TOUT	2000

/*  Definice konstant pro Loader */
/* magicke cislo udavajici platny konf. sektor */
#define MAGIC_CF_CONFIG_SEC	0x55aac396UL
/* velikost konf. oblast v sektorech */
#define LOADER_CONFSEC_SIZE	1

#define FLASH_START			FLASH_ADDRESS
#define FLASH_INT_TOP		(FLASH_ADDRESS + FLASH_SIZE - 1)
/* size of unit programmed at once */
#define FLASH_INT_BLK_SIZE	512
//#define FLASH_INT_BLK_SIZE	4
/* size of block for "Page erase" operation */
#define FLASH_INT_PAGE_ERASE_BLOCK_SIZE		2048
/* size of block for "Mass erase" operation */
#define FLASH_INT_MASS_ERASE_BLOCK_SIZE		(256*1024)
/* adresa pro ulozeni struktury citlivych informaci do int. flash */
#define FLASH_INT_SAFE_INFO_ADDR	((unsigned long)__FLASH + __FLASH_SIZE - FLASH_INT_PAGE_ERASE_BLOCK_SIZE)

/* chybove kody pro komunikaci s PC */
#define cLOADER_PCERR_LENGTH	-1	/* chybna delka prijateho paketu */
#define cLOADER_PCERR_CRC	-2	/* chyba CRC */
#define cLOADER_PCERR_FATFULL	-3	/* tabulka FAT plna */
#define cLOADER_PCERR_DATAFULL	-4	/* datova oblast plna */

/* velikost paketu pri posilani souboru
 * musi byt nasobek velikosti sektoru CF,
 * abych nemusel resit zalamovani sektoru
*/
#define cLOADER_FILE_PAKET_SIZE	1024

/* navratove kody funkci */
#define LOADER_RETOK		 	0	/* vse v poradku */
#define LOADER_RET_ERR_ERASE	-1	/* chyba pri mazani FLASH */
#define LOADER_RET_ERR_SN		-1	/* neplatne seriove cislo */
#define LOADER_ERR_CONF			-2	/* neplatny konf. sektor */
#define LOADER_ERR_CF_READ		-3	/* chyba pri cteni z CF */
#define LOADER_ERR_CF_WRITE		-4	/* chyba pri zapisudo CF */
#define LOADER_ERR_FORMAT       -5	/* chyba pri formatovani */
#define LOADER_ERR_TOUT       	-6	/* timeout pri cekani na znak */
#define LOADER_ERR_LENGTH		-7	/* paket ma spatnou delku */
#define LOADER_ERR_CRC			-8	/* neplatny kontrolni soucet paketu */
#define LOADER_ERR_LOADALL		-9	/* chyba pri nahravani souboroveho systemu */
#define LOADER_ERR_LOADAPP		-10	/* chyba pri pridavani souboru */
#define LOADER_ERR_NOTFOUND		-11	/* soubor nenalezen - hleda se ve FAT2 */
#define LOADER_ERR_IDE_WRITE	-12	/* user zapis do nepovolene oblasti CF */

/* note: external Flash memory not present for now */
#define FLASH_EXT_START	0x800000L
#define FLASH_TOP		0xC00000L
#define SLOVO_EXT_LEN	2

#define RAM_EXT_START	SDRAM_ADDRESS
#define RAM_EXT_TOP		(SDRAM_ADDRESS + SDRAM_SIZE - 1)

/* prikazy pro praci s interni FLASH */
#define IFLASH_CMD_VERIFY_ALL	0x05	/* Erase Verify 256kB */
#define IFLASH_CMD_PROGRAM		0x20	/* Program 32-bit longword */
#define IFLASH_CMD_ERASE_PAGE	0x40	/* Erase 2kB Page */
#define IFLASH_CMD_ERASE_MASS	0x41	/* Erase 256kB block */
#define IFLASH_CMD_VERIFY_PAGE	0x06	/* Erase Verify 2kB Page */

#define LOADER_HDR_OFFSET		(ETH_HDR_SIZE + IP_HDR_SIZE + UDP_HDR_SIZE)
/* UDP port vyhrazeny pro LOADER */
#define LOADER_UDP_PORT		10000

/* pocet FAT polozek v sektoru CF */
#define FATITEM_PER_SECTOR	(ATA_BLOCKSIZE/sizeof(tsimple_FAT))

/* DEFINICE MAKER */
/* prevede hodnotu v bytech na pocet sektoru CF */
#define LOADER_BYTE2SEC(x)	(((x) + ATA_BLOCKSIZE - 1)/ATA_BLOCKSIZE)

/* ceka na moznost provedeni dalsi operace Erase,Program,Verify */
#define IFLASH_WAIT_FOR_EPV_RDY	\
 while(!(MCF5282_CFM_CFMUSTAT & MCF5282_CFM_CFMUSTAT_CCIF))

/* nastavi prikaz pro Erase,Program,Verify */
#define IFLASH_SET_CMD(cmd)	do {	\
 MCF5282_CFM_CFMCMD = MCF5282_CFM_CFMCMD_CMD(cmd);	\
 } while (0)

/* spusti operaci podle prikazu v CFMCMD */
#define IFLASH_START_CMD	do {	\
 MCF5282_CFM_CFMUSTAT = MCF5282_CFM_CFMUSTAT_CBEIF;	\
 } while(0)

/* DEFINICE TYPU */
/* format Compact Flash "loader konfiguracniho sektoru"
 * fyzicky je ulozen v prvnim sektoru Compact Flash
*/
typedef struct tCF_config_sector {
    unsigned long	magic;		/* magicke cislo - oznacuje platny konf. sektor */
    unsigned long	crc;		/* kontrolni soucet konf. sektoru */
    unsigned short	version;	/* verze konf. sektoru */
    unsigned short	CF_sector_size;	/* velikost Compact Flash sektoru v bytech */
    unsigned long	CF_num_bytes;	/* velikost Compact Flash v bytech */
    unsigned long	CF_num_secs;	/* velikost Compact Flash v sektorech */
    unsigned short	FAT1_version;	/* verze FAT1 tabulky */
    unsigned long	FAT1_start;	/* prvni sektor FAT1 tabulky */
    unsigned long	FAT1_size;	/* pocet sektoru FAT1 tabulky */
    unsigned long	data1_start;	/* prvni sektor datove oblasti 1 */
    unsigned long	data1_size;	/* pocet sektoru datove oblasti 1 */
    unsigned short	FAT2_version;	/* verze FAT2 tabulky */
    unsigned long	FAT2_start;	/* prvni sektor FAT2 tabulky */
    unsigned long	FAT2_size;	/* pocet sektoru FAT2 tabulky */
    unsigned long	data2_start;	/* prvni sektor datove oblasti 2 */
    unsigned long	data2_size;	/* pocet sektoru datove oblasti 2 */
    unsigned char	CF_serial[20];	/* S/N Compact Flash - zjisteno pomoci prikazu "Identify Drive"*/
    unsigned char	TS02_serial[20];	/* S/N TS02 */
    unsigned char	eth_mac[6];	/* MAC adresa ethernetoveho rozhrani */
    unsigned char	res1[LOADER_CONFSEC_SIZE*ATA_BLOCKSIZE - 102];	/* reserved area */
} tCF_config_sector;

typedef tCF_config_sector *tpCF_config_sector;

/* struktura datoveho bloku pro prikaz "Formatovani CF" */
typedef struct tCF_format_data {
    unsigned short	FAT1_version;	/* verze FAT1 tabulky */
    unsigned long	FAT1_start;	/* prvni sektor FAT1 tabulky */
    unsigned long	FAT1_size;	/* pocet sektoru FAT1 tabulky */
    unsigned long	data1_start;	/* prvni sektor datove oblasti 1 */
    unsigned long	data1_size;	/* pocet sektoru datove oblasti 1 */
    unsigned short	FAT2_version;	/* verze FAT2 tabulky */
    unsigned long	FAT2_start;	/* prvni sektor FAT2 tabulky */
    unsigned long	FAT2_size;	/* pocet sektoru FAT2 tabulky */
    unsigned long	data2_start;	/* prvni sektor datove oblasti 2 */
    unsigned long	data2_size;	/* pocet sektoru datove oblasti 2 */
} tCF_format_data;

typedef tCF_format_data *tpCF_format_data;

/* format jednoduche FAT tabulky pro ulozeni modulu ve formatu
 * jednotlivych souboru
 * prvni volna polozka FAT tabulky musi mit 1.znak souboru roven nule 
 * delka FAT polozky musi byt volena tak (pro jednoduchost), aby se do jednoho
 * sektoru CF vesel cely pocet FAT polozek - nyni je velikost FAT
 * polozky rovna 32 bytum
*/
typedef struct tsimple_FAT {
    unsigned char	name[14];	/* nazev souboru  */
    unsigned long	size;		/* velikost souboru v bytech */
/* pocatecni sektor - offset vzhledem k prvnimu sektoru datove oblasti */
    unsigned long	start_sec;
/* koncnovy sektor - offset vzhledem k prvnimu sektoru datove oblasti */
    unsigned long	end_sec;
    unsigned short	attrib;		/* atributy souboru */
    unsigned long	crc;		/* kontrolni soucet */
} tsimple_FAT;

typedef tsimple_FAT *tpsimple_FAT;

/* struktura pro ulozeni serioveho cisla a kontrolni sumy do interni FLASH */
typedef struct tsafe_info {
    unsigned char	TS02_serial[20];	/* S/N TS02 */
    unsigned long	crc;			/* kontrolni soucet */
} tsafe_info;

typedef tsafe_info *tpsafe_info;

/* struktura pro ulozeni vstupnich bodu fci
 * zpristupnenych aplikaci (hlavnimu modulu)
*/
typedef struct tfunc4main_table {
    int (*get_hw_info)(char *buf);	/* vraci informace o HW konfiguraci */
    int (*ide_read_sector)(char *buf, unsigned long start, unsigned long num);	/* cte sektor z Compact Flash */
    int (*ide_write_sector)(char *buf, unsigned long start, unsigned long num);	/* zapise sektor do Compact Flash */
    int (*get_ip_addr)(char *ip);	/* vraci IP adresu */
    int (*set_ip_addr)(char *ip);	/* nastavi IP adresu */
    int (*get_serial_num)(char *sn);	/* vraci seriove cislo */
    int (*check_secure)(void);
    int (*loader_start)(void);		/* spusti loader bez timeoutu */
    void (*reset_system)(void);		/* provede reset systemu */
    int (*get_run_first)(int erase);	/* zjistuje, byl-li nahran novy SW */
} tfunc4main_table;

typedef tfunc4main_table *tpfunc4main_table;

#define pFM		((tpfunc4main_table)0x448)

/* operace, ktere umi Loader provest */
typedef enum tloader_operation {
  cLOADER_STBY = 0,			/* neprovadi se nic */
  cLOADER_STANDARD,			/* standardni verze loaderu */
  cLOADER_CF_INFO,			/* informace o Compact Flash */
  cLOADER_CF_FORMAT,			/* formatovani Compact Flash */
  cLOADER_CF_LOAD_ALL,			/* nahrava cely souborovy system */
  cLOADER_CF_LOAD_APPEND,		/* pridava na konec souboroveho systemu */
  cLOADER_BIOS_FLASH,			/* "flesuje" BIOS */
  cLOADER_AUTH_REQUEST,			/* zjistuje pozadavek na autorizaci */
  cLOADER_AUTH_SEND,			/* zasila autorizacni kod */
  cLOADER_LAST_ITEM
} tloader_operation;

/* format obsluzne funkce loaderu */
typedef int (*tloader_operation_fc)(void);

/* struktura pro ulozeni informaci o programovacim algoritmu */
typedef struct tCMF_prog_alg {
	unsigned int	max_pulses;		// max. pocet programovacich pulsu
	unsigned int	NVR;			// hodnota bitu CMFTST.NVR (negative voltage range)
	unsigned int	PAWS;			// hodnota bitu CMFTST.PAWS (program amplitude/width modulation select)
	unsigned int	CLKPM;			// hodnota preddelice pro generovani delky programovaciho pulsu
} tCMF_prog_alg;

/* DEFINICE EXTERNICH PROMENNYCH */
extern NIF fec_nif;

extern const unsigned char mac_default[6];

/* ukazatel na strukturu v interni FLASH,
 * kde je ulozeno ser. cislo
*/
extern volatile tpsafe_info psafe_info;

/* priznak S/N inicializovano */
extern int	fl_serial_initialized;

/* konfiguracni sektor */
extern tCF_config_sector	CF_config_sector;

/* DEFINICE EXTERNICH FUNKCI */
/*****/
/* nacte stav klavesnice
 * "posbira" stav klaves
 * z jednotlivych GPIO pinu
 * procesoru a vysledek zformuje
 * do rozumne podoby
*/
int
key_port_read(void);

/*****/
/* hleda soubor ve FAT2
 * vraci zpet prvni sektor souboru v CF - pocita se absolutne,
 * aby nemusela nadrizena aplikace hledat v konf. souborech
 * a delku souboru v bytech
 * vraci: LOADER_RETOK        - bez chyby
 *        LOADER_ERR_NOTFOUND - soubor nenalezen
*/
int
loader_find_file(
	char	*name,		/* jmeno souboru */
	uint32	*sector,	/* prvni sektor souboru */
	uint32	*size);		/* velikost souboru v bytech */

/*****/
/* nacte konfiguracni oblast CF do RAM */
/* vraci: LOADER_RETOK */
int
load_conf_sector(void);

/*****/
/* ulozi modifikovany konf. sektor do Compact Flash */
/* vraci: LOADER_RETOK */
int
save_conf_sector(void);

/*****/
/* kontroluje konf. sektor ve FLASH
 * vraci: LOADER_RETOK    - sektor platny
 *        LOADER_ERR_CONF - neplatna konfigurace
*/
int
check_conf_sector(void);

/* vypocet CRC32 pro zabezpeceni prenosu a datovych prenosu */
unsigned long
crc32 (unsigned long, const unsigned char *, unsigned int);

int
printf (const char *fmt, ...);

// funkce pro obsluhu Interni flash
int FeraseA_INT(unsigned int In_flash);
typedef int (*tFeraseA_INT)(unsigned int In_flash);
int Fwrite_INT(long adr, unsigned long *buffer_64byte,unsigned int In_flash);
typedef int (*tFwrite_INT)(long adr, unsigned long *buffer_64byte,unsigned int In_flash);
int Fverify_INT(long adr, unsigned long *buffer_64byte);

// funkce pro obsluhu Externi flash
int Ferase_EXT(int word);
int Ferase_EXT_WF( int word );
int Fwrite_EXT(long adr, unsigned char cislo);

// Zapis do externi RAM
int RAMwrite_EXT(long adresa, unsigned char znak );

/*/ fce pro komunikaci pres ethernet */
void
loader_udp_handler (NIF *nif, NBUF *pNbuf, int hdr_offset);

/*****/
int
loader_udp_IsChar(void);

/* prijme znak pomoci UDP paketu */
char
loader_udp_recv(void);

/*****/
/* vysle znak pomoci UDP paketu */
int
loader_udp_send(char znak);

// funkce pro linku, OK
int InitCom();  // tvrde na 38400
int uart_send(char zn);
char uart_recv();
int uart_IsChar();

/* fce pro komunikaci loaderu */
int send(char zn);
char recv();
int IsChar();

// Pomocn funkce
int copy(char *s,char *d, unsigned int ln);
void VysliLong (long b);
void Vyslitext(char *t);

