/****************************************************************/
/*                   Downloader do FLASH                        */
/*                    Motorola ColdFire MFC5282                 */
/*  Lubos Jelnek+Jan Damborsky                                 */
/*  vytvoreno: 12.5.2004                                        */
/*  posl. zmeny: 5.8.2004                                      */
/*                                                              */  
/*  Nahrv do FLASH, lze na stavit ku slova 1-4 byte     */  
/*  Pouziva UART1                                               */  
/*  Sedi v RAM a nahrava do interni a externi FLASH             */
/*  Spousti aplikaci v interni FLASH na druhem sektoru = 0x40000*/
/*  Rychly loader echuje pouze sumu                             */
/****************************************************************/
//*
//*  Verze pro FLASH SuperFlash SST
//*   Pouze DownLoad, a rts!!
//*
//*   Nahrava po bytech soubor Motorola Bin
//*
//*   Nejdrive kontroluje Serial number !!
//*
//*   Dela Chip erase, nebo sector erase
//*   Dela Echo kazdeho znaku v Mot file, aby brzdil prenos!!
//*
//*   Nejprve ceka na $12, vysilany periodicky z PC
//*    pak posle 'xxxx' a ceka na $21, ktery echuje
//*    a prvni radek s Serial_number
//*   Kdyz OK,posle $22, maze a posila $23
//*   Pak Mot file
//*   Radek = Delka,Adresa,Data,Suma  //Binary
//*   Pokud Adresa=-1 konec zaznamu
//*
//****************************************************************
//****************************************************************

#include "mcf5xxx.h"
#include "target.h"
#include "mcf5282.h"
#include "ip.h"
#include "nbuf.h"
#include "nif.h"
#include "b002_ide.h"
#include "b002_per.h"
#include "loader.h"

#include <string.h>

#if 1
#define	LOADER_PRINTF(fmt,args...) do { \
 if (loader_recv != uart_recv)		\
	printf(fmt ,##args);		\
} while(0)
#else
#define LOADER_PRINTF(fmt,args...)
#endif

/* DEFINICE GLOBALNICH PROMENNYCH */
/* konfiguracni oblast CF */
tCF_config_sector	CF_config_sector;

/* definice funkci pro prijem a vyslani znaku loaderu */
extern int (*loader_send)(char znak);
extern char (*loader_recv)(void);
extern int (*loader_IsChar)(void);

/* Buffer pro zapis do INT a CF Flash,
 * alokuje dostatecnou velikost pro oboje
 */
#if (FLASH_CF_BLK_SIZE > FLASH_INT_BLK_SIZE)
unsigned char FL_buff[FLASH_CF_BLK_SIZE+4];
unsigned char FL_buff2[FLASH_CF_BLK_SIZE+4];
#else
unsigned char FL_buff[FLASH_INT_BLK_SIZE+4];
unsigned char FL_buff2[FLASH_CF_BLK_SIZE+4];
#endif

/* buffer pro pousteni funkci v RAM*/
char RAM_buf[1024];

/* Definuje seriove cislo */
volatile char *serial="TS02-300000192\0";

/* definuje Magic pro detekci, jestli je pristupna FLASH */
#define     MAGIC_FLASH 0x55555555L

const char *flash_test = "UUUUUUUU\0";

/* DEFINICE EXTERNICH PROMENNCYH */
extern IP_ADDR server;

/* DEFINICE HLAVICEK LOKALNICH FUNKCI */
/*****/
/* formatuje FAT tabulku
 * vymaze vsechny soubory
 * vraci: LOADER_RETOK
 *        LOADER_ERR_FORMAT - chyba pri formatovani FAT
*/
static int
loader_fat_format(
  uint32	start,	/* poc. sektor FAT */
  uint32	size);	/* pocet sektoru FAT */

/*****/
/* nacte paket dat zadane delky
 * pokud neni pozadovany pocet znaku
 * k dispozici, vyprsi timeout
 * v promenne ln vraci pocet nactenych znaku
 * vraci: LOADER_RETOK
 *        LOADER_ERR_TOUT - vyprsel timeout pri cekani na znak
*/
static int
loader_recv_paket(
  uint8		*buff,
  uint32	*ln);

/*****/
/* nacte paket dat zadane delky
 * [1] pokud neni pozadovany pocet znaku
 *     k dispozici, vyprsi timeout
 *     v promenne ln vraci pocet nactenych znaku
 * [2] provede kontrolu CRC, pokud je v poradku
 *     posle CRC zpet - 4 byty
 * [3] pokud doslo k chybe, posle zpet kod chyby - 2 byty
 *
 * vraci: LOADER_RETOK
 *        LOADER_ERR_TOUT   - vyprsel timeout pri cekani na znak
 *        LOADER_ERR_LENGTH - v bufferu zustaly neprectene znaky
 *        LOADER_ERR_CRC    - chyba kontrolniho souctu
*/
static int
loader_recvchk_paket(
  uint8		*buff,
  uint32	*ln,
  int		ack);

/*****/
/* vysle paket dat.
 * pokud vysila na ser. linku, posle 
 * paket po znacich,
 * pokud komunikuje pres UDP,
 * rozseka data na UDP pakety
*/
static int
loader_send_paket(
  uint8		*buff,
  uint32	ln);

/*****/
/* standardni verze loaderu */
static int
loader_standard(void);

/*****/
/* informace o Compact Flash */
static int
loader_cf_info(void);

/*****/
/* formatovani Compact Flash
 * plni polozky CF_config_sector.
 * vraci: LOADER_RETOK        - v poradku
 *        LOADER_ERR_FORMAT   - chyba pri formatovani
 *        LOADER_ERR_CF_WRITE - chyba pri zapisu konf. sektoru
 */
static int
loader_cf_format(void);

/*****/
/* nahrava cely souborovy system */
int
loader_cf_load_all(void);

/*****/
/* pridava na konec souboroveho systemu */
int
loader_cf_load_append(void);

/*****/
/* "flesuje" BIOS */
int
loader_bios_flash(void);

/*****/
/* zjistuje pozadavek na autorizaci */
int
loader_auth_request(void);

/*****/
/* zasila autorizacni kod */
int
loader_auth_send(void);

/* DEFINICE STATICKYCH PROMENNYCH */
#define cLOADER_RXBUF_LN	2000
static int recv_char[cLOADER_RXBUF_LN];	 /* buffer znaku prijaty pomoci UDP paketu */
static int recv_char_fl = FALSE; /* pokud TRUE, byl prijat platny znak */
static int recv_char_ln = 0;	 /* pocet prijatych dat v bufferu */
static int recv_char_head = 0;	 /* index akt. zpracovavaneho znaku */
static int recv_char_tail = 0;   /* index akt. prijateho znaku */

#ifdef FAST_LOADER
static int loader_fast_fl = FALSE;	/* == TRUE..vyuziva rychleho protokolu */
#endif

/* aktualne provadena operace, default je nahrani souboroveho systemu */
static tloader_operation	loader_operation = cLOADER_STBY;

/* tabulka fnci pro loader */
static tloader_operation_fc	loader_operation_table[cLOADER_LAST_ITEM] = {
  NULL,				/* neprovadi se nic */
  loader_standard,		/* standardni verze loaderu */
  loader_cf_info,		/* informace o Compact Flash */
  loader_cf_format,		/* formatovani Compact Flash */
  loader_cf_load_all,		/* nahrava cely souborovy system */
  loader_cf_load_append,	/* pridava na konec souboroveho systemu */
  loader_bios_flash,		/* "flesuje" BIOS */
  loader_auth_request,		/* zjistuje pozadavek na autorizaci */
  loader_auth_send,		/* zasila autorizacni kod */
};

/* DEFINICE LOKALNICH FUNKCI */
/*****/
/* formatuje FAT tabulku
 * vymaze vsechny soubory
 * vraci: LOADER_RETOK
 *        LOADER_ERR_FORMAT - chyba pri formatovani FAT
*/
static int
loader_fat_format(
  uint32	start,	/* poc. sektor FAT */
  uint32	size)	/* pocet sektoru FAT */
{
    int		i,j;
    /* sektor FAT tabulky */
    tsimple_FAT	FAT_sec[FATITEM_PER_SECTOR];

    LOADER_PRINTF(" format <%ld,%ld>..",(long)start, (long)(start + size - 1));

    for(i = 0; i < FATITEM_PER_SECTOR; i++) {
	for(j = 0; j < sizeof(FAT_sec[i].name); j++)
	    FAT_sec[i].name[j] = '\0';
    } /*for*/

    for(i = 0; i < size; i++) {
	if (ide_write(0, start + i,  1, (uint32 *)FAT_sec) != 1) {
	    LOADER_PRINTF("\nFAT format: Error when writting sektor %ld",
	     (long)(start + i));

	    return(LOADER_ERR_FORMAT);
	} /*endif*/
    } /*for*/

    return(LOADER_RETOK);
} /* end of loader_fat_format() */

/*****/
/* standardni verze loaderu */
static int
loader_standard(void)
{
    return(LOADER_RETOK);
} /* end of loader_standard() */

/*****/
/* informace o Compact Flash */
static int
loader_cf_info(void)
{
    uint8	cf_info[LOADER_CONFSEC_SIZE*ATA_BLOCKSIZE + 4];
    uint8	*dp, *sp;
    int		i;

    /* zkopiruje konfiguracni parametry do bufferu */
    for(i = 0, dp = cf_info, sp = (uint8 *)&CF_config_sector;
     i < sizeof(CF_config_sector); i++)
	*dp++ = *sp++;

    /* spocita zabezpeni pro prenos */
    *((uint32 *)(cf_info + LOADER_CONFSEC_SIZE*ATA_BLOCKSIZE))
     = crc32(0, cf_info, LOADER_CONFSEC_SIZE*ATA_BLOCKSIZE);

    /* posle paket do PC */
    return(loader_send_paket(cf_info, sizeof(cf_info)));
} /* end of loader_cf_info() */

/*****/
/* formatovani Compact Flash
 * plni polozky CF_config_sector.
 * vraci: LOADER_RETOK        - v poradku
 *        LOADER_ERR_FORMAT   - chyba pri formatovani
 *        LOADER_ERR_CF_WRITE - chyba pri zapisu konf. sektoru
 */
static int
loader_cf_format(void)
{
    uint8	buff[sizeof(tCF_format_data) + sizeof(uint32)];
    int		i;
    uint8	*dp,*sp;
    uint32	ln = sizeof(tCF_format_data);

    /* prijmu data pro format a crc32 */
    if (loader_recvchk_paket(buff, &ln, FALSE) != LOADER_RETOK)
	return(LOADER_ERR_FORMAT);

    /* zkopiruje nova data */
    dp = (uint8 *)&CF_config_sector.FAT1_version;
    sp = buff;
    for (i = 0; i < sizeof(tCF_format_data); i++)
	*dp++ = *sp++;

    /* ulozi konfiguracni sektor */
    save_conf_sector();

    /* zformatuje FAT1, FAT2 tabulku
     * naplni vsechny FAT polozky name
     * nulami
    */
    /* format FAT1 */
    LOADER_PRINTF("\nCF format: Formatting FAT1..");
    if (loader_fat_format(CF_config_sector.FAT1_start, CF_config_sector.FAT1_size) != LOADER_RETOK)
	return(LOADER_ERR_FORMAT);

    /* format FAT2 */
    LOADER_PRINTF("OK\nCF format: Formatting FAT2..");
    if (loader_fat_format(CF_config_sector.FAT2_start, CF_config_sector.FAT2_size) != LOADER_RETOK)
	return(LOADER_ERR_FORMAT);

    LOADER_PRINTF("OK");

    /* nacte konfiguracni sektor */
    load_conf_sector();

    return(loader_cf_info());
} /* end of loader_cf_format() */

/*****/
/* nahrava cely souborovy system
 * [1] prijme obsah FAT polozky, odpovi CRC
 * [2] prijme soubor v binarnim tvaru po paketech, odpovida
 *     po kazdem paketu CRC
 * [3] opakuje [1] a [2], dokud nejsou prijaty vsechny soubory
 * [4] ukonci prenos, pokud polozka name FAT tabulky je vyplnena
 *     nulami
 * Pokud behem prenosu dojde k chybe delky paketu nebo CRC,
 * ukonci prenos a posle do PC chybovy kod
 * 
 * vraci: LOADER_RETOK
 *        LOADER_ERR_LOADALL  - chyba pri nahravani souboroveho systemu
*/
int
loader_cf_load_all(void)
{
    uint8		buff[1500];
    tsimple_FAT		FAT_item;
    int32		ln;
    /* aktualni cislo sektoru FAT tabulky */
    uint32		FAT_current_sec_num;
    /* aktualni sektor FAT tabulky */
    tsimple_FAT		FAT_current_sec[FATITEM_PER_SECTOR];
    /* aktualni polozka FAT tabulky v aktualnim sektoru */
    uint32		FAT_current_item;
    /* aktualni sektor pro ulozeni souboru */
    uint32		FILE_current_sec_num;
    char		file_name[15];
    int			i;

    /* vyzvedne informace o ulozeni FAT tab a datove casti */
    FAT_current_sec_num = CF_config_sector.FAT2_start;
    FAT_current_item = 0;
    FILE_current_sec_num = CF_config_sector.data2_start;

    /* format FAT2 */
    LOADER_PRINTF("\nCF format: Formatting FAT2..");
    if (loader_fat_format(CF_config_sector.FAT2_start, CF_config_sector.FAT2_size) != LOADER_RETOK)
	return(LOADER_ERR_LOADALL);

    LOADER_PRINTF("OK");

    /* smaze pracovni kopii FAT sektoru v RAM
     * maze pouze prvni ctyri znaky nazvu souboru
     * v kazde polozce FAT sektoru
    */
    for(i = 0; i < FATITEM_PER_SECTOR; i++)
	*((uint32 *)FAT_current_sec[i].name) = 0;

    do {
	/* prijme polozku FAT tabulky */
	ln = sizeof(tsimple_FAT);

	if (loader_recvchk_paket(buff, &ln, TRUE) != LOADER_RETOK)
	    return(LOADER_ERR_LOADALL);

	/* zkopiruje z prijimaciho bufferu FAT polozku */
	FAT_item = *((tpsimple_FAT)buff);

	/* pokud je polozka name vyplnena nulami, konec ukladani */
	/* preskoci na dalsi fazi */
	if (!FAT_item.name[0]) {
	    LOADER_PRINTF("\nCF LoadAll: Files loaded");

	    /* ulozi posledne rozpracovany sektor FAT tabulky */
	    if (FAT_current_item) {
		if (ide_write(0, FAT_current_sec_num, 1, (uint32 *)FAT_current_sec) != 1)
		    LOADER_PRINTF("\nCF LoadAll: Error when writing FAT sektor %ld", (long)FAT_current_sec);
	    } /*endif*/

	    break;
	} /*endif*/

	/* vyplni ve FAT polozce poc. a koncovy sektor souboru
	 * jsou to offsety vzhledem k poc. sektoru dat. Nejdrive
	 * zkontroluje, jestli je v datove oblasti dostatek mista
	 * pro soubor.
	*/
	if ((FILE_current_sec_num - CF_config_sector.data2_start + LOADER_BYTE2SEC(FAT_item.size))
	 > CF_config_sector.data2_size) {
		LOADER_PRINTF("\nCF LoadAll: data2 full");

		*((uint16 *)buff) = cLOADER_PCERR_DATAFULL;
		loader_send_paket(buff, sizeof(uint16));

		return(LOADER_ERR_LOADALL);
	} /*endif*/
	FAT_item.start_sec = FILE_current_sec_num - CF_config_sector.data2_start;
	FAT_item.end_sec = FAT_item.start_sec + LOADER_BYTE2SEC(FAT_item.size) - 1;

	/* pokud neni ve FAT misto, konec */
	if (!FAT_current_item
	 && (FAT_current_sec_num - CF_config_sector.FAT2_start) >= CF_config_sector.FAT2_size) {
	    LOADER_PRINTF("\nCF LoadAll: FAT2 full");

	    *((uint16 *)buff) = cLOADER_PCERR_FATFULL;

	    loader_send_paket(buff, sizeof(uint16));

	    return(LOADER_ERR_LOADALL);
	} /*endif*/

	/* vypise info o ukladanem souboru */
	for(i = 0; i < sizeof(FAT_item.name) && FAT_item.name[i]; i++)
	    file_name[i] = FAT_item.name[i];
	file_name[i] = '\0';
	LOADER_PRINTF("\nFile %s: size=%ld, attr=%04lX, sec<%ld;%ld>", file_name,
	 (long)FAT_item.size, (long)FAT_item.attrib, (long)FAT_item.start_sec, (long)FAT_item.end_sec);
      
	/* ulozi data do FAT tabulky */
	FAT_current_sec[FAT_current_item++] = FAT_item;

	/* pokud je zaplnen cely sektor, uloz ho a smaz kopii v RAM */
	if (FAT_current_item >= sizeof(FAT_current_sec)/sizeof(tsimple_FAT)) {
	    FAT_current_item = 0;

	    if (ide_write(0, FAT_current_sec_num, 1, (uint32 *)FAT_current_sec) != 1)
		LOADER_PRINTF("\nCF LoadAll: Error when writing FAT sektor %ld", (long)FAT_current_sec);

	    /* smaz pracovni kopii v RAM */
	    for(i = 0; i < FATITEM_PER_SECTOR; i++)
		*((uint32 *)FAT_current_sec[i].name) = 0;

	    FAT_current_sec_num++;
	} /*endif*/

	/* nacte obraz souboru a uklada ho do Compact Flash
	 * soubor se posila po paketech. V kazdem paketu
	 * je cLOADER_FILE_PAKET_SIZE bytu, posledni paket
	 * je kratsi. Kazdy paket je zabezpecen CRC, ktery
	 * se echuje zpet. Pokud dojde behem prenosu k chybe,
	 * nahravani se prerusi a posle se zpet kod chyby (2 byty)
	*/
	for(ln = FAT_item.size; ln > 0; ln -= cLOADER_FILE_PAKET_SIZE) {
	    uint32 l = (ln >= cLOADER_FILE_PAKET_SIZE) ? cLOADER_FILE_PAKET_SIZE : ln;
	    uint32 num_sec = LOADER_BYTE2SEC(l);

	    /* nacte paket */
	    if (loader_recvchk_paket(buff, &l, TRUE) != LOADER_RETOK)
		return(LOADER_ERR_LOADALL);
	    
	    LOADER_PRINTF("%s", "#");

	    /* uloz prijata data do CF */
	    if (ide_write(0, FILE_current_sec_num, num_sec, (uint32 *)buff) != num_sec) {
		LOADER_PRINTF("\nCF LoadAll: Error when writing FAT sektor %ld", (long)FAT_current_sec);

		return(LOADER_ERR_LOADALL);
	    } /*endif*/

	    /* posune se na dalsi volny sektor */
	    FILE_current_sec_num += num_sec;
	} /*for*/

	LOADER_PRINTF("%s", "done");
    } while(TRUE);

    return(LOADER_RETOK);
} /* end of loader_cf_load_all() */

/*****/
/* pridava jeden soubor na konec souboroveho systemu
 * format nahravani viz loader_cf_load_all()
 * vraci: LOADER_RETOK       - v poradku
 *        LOADER_ERR_LOADAPP - chyba
*/
int
loader_cf_load_append(void)
{
    uint8		buff[1500];
    tsimple_FAT		FAT_item;
    /* posledni pouzity datovy sektor */
    uint32		first_free_sector;
    int32		ln;
    /* aktualni cislo sektoru FAT tabulky */
    uint32		FAT_current_sec_num;
    /* aktualni sektor FAT tabulky */
    tsimple_FAT		FAT_current_sec[FATITEM_PER_SECTOR];
    /* aktualni cislo polozky FAT tabulky v aktualnim sektoru */
    uint32		FAT_current_item;
    char		file_name[15];
    int			i;

    /* vyzvedne informace o ulozeni FAT tab a datove casti */
    FAT_current_sec_num = CF_config_sector.FAT2_start;

    /* posledne pouzity datovy sektor predpoklada 1. volny
     * Nutno inicializovat pro pripad, ze jeste nebyl nahran
     * zadny soubor a pridava se na pocatek datove oblasti
    */
    first_free_sector = CF_config_sector.data2_start;

    LOADER_PRINTF("\nCF LoadAppend: FAT start:%ld, data start:%ld",
     (long)FAT_current_sec_num, (long)first_free_sector);

    /* najde ve FAT tabulce volne misto */
    do {
	/* pokud neni ve FAT volne misto, konec */
	if (FAT_current_sec_num >= CF_config_sector.FAT2_start + CF_config_sector.FAT2_size) {
	    LOADER_PRINTF("\nCF LoadAppend: FAT2 full");

	    *((uint16 *)buff) = cLOADER_PCERR_FATFULL;

	    loader_send_paket(buff, sizeof(uint16));

	    return(LOADER_ERR_LOADAPP);
	} /*endif*/

	/* z posledni FAT polozky sektoru si vzdy ulozi cislo posledne pouziteho
	 * sektoru dat
	*/
	if (FAT_current_sec_num != CF_config_sector.FAT2_start)
	    first_free_sector = FAT_current_sec[FATITEM_PER_SECTOR - 1].end_sec + 1;

	/* nacte z CF dalsi sektor FAT tabulky */
	if (ide_read(0, FAT_current_sec_num, 1, (uint32 *)FAT_current_sec) != 1) {
	    LOADER_PRINTF("\nCF LoadAppend: Error when reading FAT sektor %ld", (long)FAT_current_sec);

	    return(LOADER_ERR_LOADAPP);
	} /*endif*/

	FAT_current_sec_num++;

	/* projde postupne FAT polozky nacteneho sektoru
	 * a hleda, jestli nenajde volnou polozku
	 * Behem hledani si vzdy ulozi z kazde FAT polozky hodnotu
	 * posledne pouziteho sektoru
	*/
	for(i = 0;
	 i < FATITEM_PER_SECTOR && FAT_current_sec[i].name[0]; i++) {
	    FAT_item = FAT_current_sec[i];

	    LOADER_PRINTF("\nFile %s: size=%ld, attr=%04lX, sec<%ld;%ld>", FAT_item.name,
	     (long)FAT_item.size, (long)FAT_item.attrib, (long)FAT_item.start_sec, (long)FAT_item.end_sec);
	} /*for*/
    } while(i == FATITEM_PER_SECTOR);

    /* pokud je prvni volna FAT polozka na zacatku sektoru, pouzije
     * obsah promenne last_used_sector prevzate z predchoziho sektoru
    */
    if (i)
	first_free_sector = FAT_current_sec[i-1].end_sec + CF_config_sector.data2_start + 1;

    FAT_current_item = i;

    LOADER_PRINTF("\nCF LoadAppend: Found free FAT item %ld in FAT sector %ld",
     (long)FAT_current_item, (long)(FAT_current_sec_num - CF_config_sector.FAT2_start));
    LOADER_PRINTF("\nCF LoadAppend: First free data sector: %ld",
     (long)(first_free_sector - CF_config_sector.data2_start));
    LOADER_PRINTF("\nCF LoadAppend: Num of free data sectors: %ld",
     (long)(CF_config_sector.data2_size - (first_free_sector - CF_config_sector.data2_start)));

    /* prijme polozku FAT tabulky */
    ln = sizeof(tsimple_FAT);
    if (loader_recvchk_paket(buff, &ln, TRUE) != LOADER_RETOK)
	return(LOADER_ERR_LOADAPP);

    /* zkopiruje z prijimaciho bufferu FAT polozku */
    FAT_item = *((tpsimple_FAT)buff);

    /* vyplni ve FAT polozce poc. a koncovy sektor souboru
     * jsou to offsety vzhledem k poc. sektoru dat. Nejdrive
     * zkontroluje, jestli je v datove oblasti dostatek mista
     * pro soubor.
    */
    if ((first_free_sector - CF_config_sector.data2_start + LOADER_BYTE2SEC(FAT_item.size))
	 > CF_config_sector.data2_size) {
	LOADER_PRINTF("\nCF LoadAppend: data2 full");

	*((uint16 *)buff) = cLOADER_PCERR_DATAFULL;
	loader_send_paket(buff, sizeof(uint16));

	return(LOADER_ERR_LOADAPP);
    } /*endif*/

    FAT_item.start_sec = first_free_sector - CF_config_sector.data2_start;
    FAT_item.end_sec = FAT_item.start_sec + LOADER_BYTE2SEC(FAT_item.size) - 1;

    /* vypise info o ukladanem souboru */
    for(i = 0; i < sizeof(FAT_item.name) && FAT_item.name[i]; i++)
        file_name[i] = FAT_item.name[i];

    file_name[i] = '\0';
    LOADER_PRINTF("\nFile %s: size=%ld, attr=%04lX, sec<%ld;%ld>", file_name,
     (long)FAT_item.size, (long)FAT_item.attrib, (long)FAT_item.start_sec, (long)FAT_item.end_sec);
      
    /* ulozi data do FAT tabulky a ulozi FAT sektor */
    FAT_current_sec[FAT_current_item] = FAT_item;

    if (ide_write(0, FAT_current_sec_num - 1, 1, (uint32 *)FAT_current_sec) != 1)
	LOADER_PRINTF("\nCF LoadAppend: Error when writing FAT sektor %ld", (long)FAT_current_sec);

    /* nacte obraz souboru a uklada ho do Compact Flash
     * soubor se posila po paketech. V kazdem paketu
     * je cLOADER_FILE_PAKET_SIZE bytu, posledni paket
     * je kratsi. Kazdy paket je zabezpecen CRC, ktery
     * se echuje zpet. Pokud dojde behem prenosu k chybe,
     * nahravani se prerusi a posle se zpet kod chyby (2 byty)
    */
    for(ln = FAT_item.size; ln > 0; ln -= cLOADER_FILE_PAKET_SIZE) {
	uint32 l = (ln >= cLOADER_FILE_PAKET_SIZE) ? cLOADER_FILE_PAKET_SIZE : ln;
	uint32 num_sec = LOADER_BYTE2SEC(l);

	/* nacte paket */
	if (loader_recvchk_paket(buff, &l, TRUE) != LOADER_RETOK)
	    return(LOADER_ERR_LOADALL);
	    
	LOADER_PRINTF("%s", "#");

	/* uloz prijata data do CF */
	if (ide_write(0, first_free_sector, num_sec, (uint32 *)buff) != num_sec) {
	    LOADER_PRINTF("\nCF LoadAll: Error when writing FAT sektor %ld", (long)first_free_sector);

	    return(LOADER_ERR_LOADALL);
	} /*endif*/

	/* posune se na dalsi volny sektor */
	first_free_sector += num_sec;
    } /*for*/

    LOADER_PRINTF("%s", "done");

    return(LOADER_RETOK);
} /* end of loader_cf_load_append() */

/*****/
/* "flesuje" BIOS */
int
loader_bios_flash(void)
{
    return(LOADER_RETOK);
} /* end of loader_bios_flash() */

/*****/
/* zjistuje pozadavek na autorizaci */
int
loader_auth_request(void)
{
    return(LOADER_RETOK);
} /* end of loader_auth_request() */

/*****/
/* zasila autorizacni kod */
int
loader_auth_send(void)
{
    return(LOADER_RETOK);
} /* end of loader_auth_send() */

/* DEFINICE GLOBALNICH FUNKCI */
/*****/
/* nacte stav klavesnice
 * "posbira" stav klaves
 * z jednotlivych GPIO pinu
 * procesoru a vysledek zformuje
 * do rozumne podoby
*/
int
key_port_read(void)
{
	int		key_port = 0;

	/* Key_[0,1] on PC[2,3] */
	key_port |= (MCF5282_GPIO_PORTCP >> 2) & 0x03;

	/* Key_2 on PD[4] */
	key_port |= (MCF5282_GPIO_PORTDP >> 2) & 0x04;

	/* Key_[3,4] on PC[6,5] */
	key_port |= (MCF5282_GPIO_PORTCP >> 2) & 0x18;

	/* Key_[5,6] on PTC[1..0] */
	key_port |= (MCF5282_GPIO_PORTTCP << 5) & 0x60;

	/* Key_[7..9] on PTD[3..1] */
	key_port |= ((int)MCF5282_GPIO_PORTTDP <<  6) & 0x380;

	return(~key_port & 0x3ff);
} /* end of read_key_port() */

/*****/
/* hleda soubor ve FAT2
 * vraci zpet prvni sektor souboru v CF - pocita se absolutne,
 * aby nemusela nadrizena aplikace hledat v konf. souborech
 * a delku souboru v bytech
 * vraci: LOADER_RETOK        - bez chyby
 *        LOADER_ERR_NOTFOUND - soubor nenalezen
*/
int
loader_find_file(
	char	*fname,		/* jmeno souboru */
	uint32	*sector,	/* prvni sektor souboru */
	uint32	*size)		/* velikost souboru v bytech */
{
	int				i, j;
    tsimple_FAT		FAT_current_sec[FATITEM_PER_SECTOR];

	/* prochazi postupne celou FAT2 tabulku po polozkach
	 *  a hleda v nazvu souboru shodu
	*/
	for(i = 0; i < CF_config_sector.FAT2_size; i++) {
		/* nacte sektor FAT tabulky - vytvori pracovni kopii v RAM */
		ide_read(0, CF_config_sector.FAT2_start + i, 1, (uint32 *)FAT_current_sec);

		/* projde vsechny FAT polozky v sektoru a hleda shodu ve jmene souboru */
		for(j = 0; j < FATITEM_PER_SECTOR; j++) {
			/* pokud jmeno souboru zacina '\0', dosel na konec FAT tabulky 
			 * a soubor nebyl nalezen, tudiz skonci
			*/
			if (!FAT_current_sec[j].name[0]) {
				LOADER_PRINTF("\nFind file:Soubor %s nenalezen", fname);

				return(LOADER_ERR_NOTFOUND);
			} /*endif*/

			if (strncmp(FAT_current_sec[j].name, fname, sizeof(FAT_current_sec[j].name)) == 0) {
				/* nalezena shoda, zkopiruje informace o souboru a konec */
				*sector = FAT_current_sec[j].start_sec + CF_config_sector.data2_start;
				*size = FAT_current_sec[j].size;
				LOADER_PRINTF("\nFind file:Soubor %s zacina v sektoru %ld, delka %ld", fname,
				 *sector, *size);

				return(LOADER_RETOK);
			} /*endif*/
		} /*for*/
	} /*for*/

	/* prosel celou FAT tabulku a nenasel soubor - konec */
	return(LOADER_ERR_NOTFOUND);
} /* end of loader_find_file() */

/*****/
/* nacte kopii konfiguracni oblasti CF do RAM
 * vraci: LOADER_RETOK
 *        LOADER_ERR_CF_READ - chyba pri cteni z CF
*/
int
load_conf_sector(void)
{
    return(ide_read(0, 0, LOADER_CONFSEC_SIZE, (uint32 *)&CF_config_sector) == LOADER_CONFSEC_SIZE ?
     LOADER_RETOK : LOADER_ERR_CF_READ);
} /* end of load_conf_sector() */

/*****/
/* ulozi modifikovany konf. sektor do Compact Flash
 * pritom opravi polozky CF_config_sector.magic,
 * CF_config_sector.crc a polozky tykajici se
 * udaju o CF
 *
 * vraci: LOADER_RETOK
 *        LOADER_ERR_CF_WRITE - chyba pri zapisu konf. sektoru do CF
*/
int
save_conf_sector(void)
{
    int	i;

    /* verze boot sektoru: 1.1 */
    CF_config_sector.version = 0x0101;

    /* upravi polozky popisujici CF */
    /* velikost Compact Flash sektoru v bytech */
    CF_config_sector.CF_sector_size = ATA_BLOCKSIZE;
    /* velikost Compact Flash v sektorech */
    CF_config_sector.CF_num_secs = ide_dev_desc[0].lba;
    /* velikost Compact Flash v bytech */
    CF_config_sector.CF_num_bytes = CF_config_sector.CF_sector_size * CF_config_sector.CF_num_secs;
    /* Compact Flash S/N */
    for(i = 0; i < sizeof(CF_config_sector.CF_serial) - 1
     && (ide_dev_desc[0].product[i]); i++)
	CF_config_sector.CF_serial[i] = ide_dev_desc[0].product[i];

    CF_config_sector.CF_serial[i] = '\0';

    /* upravi MAGIC a kontrolni soucet */
    CF_config_sector.magic = MAGIC_CF_CONFIG_SEC;

    CF_config_sector.crc = crc32(0, (uint8 *)&CF_config_sector.version,
      (uint32)CF_config_sector.eth_mac - (uint32)&CF_config_sector.version);

    return(ide_write(0, 0, LOADER_CONFSEC_SIZE, (uint32 *)&CF_config_sector) == LOADER_CONFSEC_SIZE ?
     LOADER_RETOK : LOADER_ERR_CF_WRITE);
} /* end of save_conf_sector() */

/*****/
/* kontroluje konf. sektor ve FLASH
 * zkontroluje MAGIC NUMBER a crc32 kontrolni soucet
 * vraci: LOADER_RETOK    - sektor platny
 *        LOADER_ERR_CONF - neplatna konfigurace
*/
int
check_conf_sector(void)
{
    /* nejdrive testuje magic */
    if (CF_config_sector.magic != MAGIC_CF_CONFIG_SEC)
	return(LOADER_ERR_CONF);

    /* provede kontrolu CRC32.
     * pocita se z polozek CF_config_sector.version
     * az CF_config_sector.TS02_serial
    */
    if (CF_config_sector.crc
     != crc32(0, (uint8 *)&CF_config_sector.version,
      (uint32)CF_config_sector.eth_mac - (uint32)&CF_config_sector.version))
	return(LOADER_ERR_CONF);

    return(LOADER_RETOK);
} /* end of check_conf_sector() */

/*********************************************/
/*********************************************/
/*********************************************/
/*  LOADER                                   */
/*  0 - zapisuje do cele vnitrni a vnejsi FLASH, lezi v RAM nahore */
/*  1 - zapisuje do vnitrni a vnejsi FLASH, lezi v 1.sektoru FLASH */
/*  2 - nahrava do RAM,  lezi v 1.sektoru FLASH */
int
loader(unsigned int In_flash)
{
	int 	i;
	int		r = 0;
/*	int		j; */
	int 	bytu_inline;
	char 	znak;
	unsigned int	suma,buff_cnt;
	unsigned long	adresa,x;
	unsigned long	adresa_buff = 0;
	int		fl_serial_ok;

	/* rozhodne se, jestli bude pouzivat rychly protokol
	 * pokud loaduje po ethernetu, tak jo,
         * pro ser. linku je potreba dodelat
         */
#ifdef FAST_LOADER
	if (loader_send == loader_udp_send) {
	    LOADER_PRINTF("\nFast loader");
	    loader_fast_fl = TRUE;
	} else {
	    LOADER_PRINTF("\nSlow loader");
	    loader_fast_fl = FALSE;
	}
#endif

	// tady probehne nejaky init handshake
	loader_send('T');
	loader_send('S');
	loader_send('0');
	loader_send('2');

	/* zatim nevi, co se bude provadet */
	loader_operation = cLOADER_STBY;

	do {
	    znak = loader_recv();
	    
	    /* rozhodne se podle prijateho znaku, kterou operaci bude provadet */
	    switch (znak) {
		/* standardni loader */
		case 0x21:
		    loader_operation = cLOADER_STANDARD;
	    	    LOADER_PRINTF("\nLoader OP:Standard");
		    break;

		/* nahrava se cely souborovy system */
		case 0x22:
		    loader_operation = cLOADER_CF_LOAD_ALL;
	    	    LOADER_PRINTF("\nLoader OP:CF Load All");
		    break;

		/* nahrava se soubor na konec */
		case 0x23:
		    loader_operation = cLOADER_CF_LOAD_APPEND;
	    	    LOADER_PRINTF("\nLoader OP:CF Load Append");
		    break;

		/* info o CF */
		case 0x24:
		    loader_operation = cLOADER_CF_INFO;
	    	    LOADER_PRINTF("\nLoader OP:CF Info");
		    break;

		/* formatovani CF */
		case 0x25:
		    loader_operation = cLOADER_CF_FORMAT;
	    	    LOADER_PRINTF("\nLoader OP:CF Format");
		    break;

		/* pozadavek na autorizaci */
		case 0x26:
		    loader_operation = cLOADER_AUTH_REQUEST;
	    	    LOADER_PRINTF("\nLoader OP:Authorization Request");
		    break;

		/* autorizace */
		case 0x27:
		    loader_operation = cLOADER_AUTH_SEND;
	    	    LOADER_PRINTF("\nLoader OP:Authorization Send");
		    break;
	    } /* switch */
	} while(loader_operation == cLOADER_STBY);
	
	/* echo - v1.2 misto 0x21 posila 0x31 (echuje z celeho radku pouze sumu) */
#ifdef FAST_LOADER
	if (loader_fast_fl)
	    loader_send(0x31);
	else
#endif
	    loader_send(0x21);

	/* Nactu S/N - predpokladam, ze je v poradku */
	fl_serial_ok = TRUE;
	for(i=0; i < SERIAL_COUNT; i++) {
		znak = loader_recv() & 0xff;

		/* pokud neni S/N inicializovano, neprovadi komparaci,
		 * ale ulozi ho do FLASH pameti
                 */
		if (fl_serial_initialized) {
		    if(znak != psafe_info->TS02_serial[i])
			fl_serial_ok = FALSE;
		} else {
		    serial[i] = znak;
		}
	} /*for*/

	/* je-li S/N neplatne, koncim */
	if (!fl_serial_ok)
	    return(LOADER_RET_ERR_SN);

	/* S/N v poradku */
	loader_send(0x22);

	/* pokud se nahrava poprve, ulozi ser. cislo do interni FLASH,
         * MAC adresu, do konf. sektoru Compact Flash
         */
	if (!fl_serial_initialized) {
	    unsigned char 	buff[FLASH_INT_BLK_SIZE];
	    int			x,r;

	    /* prohlasim S/N od nynejska za platne */
	    fl_serial_initialized = TRUE;

	    /* nactu obsah progr. jednotky do bufferu */
	    for(x = 0; x < (FLASH_INT_BLK_SIZE/sizeof(uint32)); x++)
		((uint32*)buff)[x] = ((uint32*)FLASH_INT_SAFE_INFO_ADDR)[x];

	    /* modifikuju cast odpovidajici S/N */
	    for(x = 0; x < SERIAL_COUNT; x++)
		((tpsafe_info)buff)->TS02_serial[x] = serial[x];

	    /* ulozi sektor do interni flash */
	    if(In_flash == 1) {
		copy((char *)Fwrite_INT, RAM_buf, sizeof(RAM_buf));
		r = ((tFwrite_INT)RAM_buf)((uint32)FLASH_INT_SAFE_INFO_ADDR, (unsigned long *)buff, 1);
	    } else
		r = Fwrite_INT((uint32)FLASH_INT_SAFE_INFO_ADDR, (unsigned long *)buff, 0);

	    /* ulozi seriove cislo */
	    for(i = 0; i < SERIAL_COUNT; i++)
		CF_config_sector.TS02_serial[i] = serial[i];

	    CF_config_sector.TS02_serial[SERIAL_COUNT] = '\0';

	    /* sestavi ze ser. cisla MAC adresu */
	    CF_config_sector.eth_mac[0] = mac_default[0];
	    CF_config_sector.eth_mac[1] = mac_default[1];
	    CF_config_sector.eth_mac[2] = mac_default[2];
	    CF_config_sector.eth_mac[3] = mac_default[3];
	    CF_config_sector.eth_mac[4] = serial[12];
	    CF_config_sector.eth_mac[5] = serial[13];

	    /* ulozi modifikovany konf. sektor */
	    save_conf_sector();
	} /*endif*/

	/* zavola obsluznou rutinu pro danou fci
	 * nebo spusti standardni verzi loaderu
	 * TODO: presunout standardni loader do samostatne fce
	*/
	if (loader_operation != cLOADER_STANDARD) {
	    loader_operation_table[loader_operation]();
	    return(LOADER_RETOK);
	} /*endif*/

	/* Mazu INT a EXT Flash, pokud v FLASH, tak kopiruju do RWM */
	/* Int. FLASH mazu pouze tehdy, pokud Loader sedi na zacatku FLASH */
	if ((In_flash == 1) && (__FLASH == __TEXT_START)) {
		copy((char *)FeraseA_INT,RAM_buf, sizeof(RAM_buf));
		
		// spustim funkci ktera je a zacina v RAM
		r = ((tFeraseA_INT)RAM_buf)(1);
	}

	if(In_flash == 0)
	{
		/* maze celou interni flash */
		r = FeraseA_INT(0);
/*		r = FeraseA_INT(1); */
	}

	if(r < 0)
	    return(-3);

	for(x=0; x<5000000; x++);
	/* pamet smazana */
	loader_send(0x23);

/*	Fwrite_INT(0, &adresa_buff, 0); */

	// ted si nakopiruju zapis do RAM, aby mne to pak nezdrzovalo ! 
	if(In_flash == 1)
		copy((char *)Fwrite_INT, RAM_buf, sizeof(RAM_buf));

	/* pocitam byte do FLASH_INT_BLK_SIZE do Interni Flash */
	buff_cnt = -1;
	
	/* samotny download */
	do {
		/* nactu pocet bytu, minus delka adresy a sumy */
		znak = loader_recv();

		/* echo ve verzi v1.2 nedela */
#ifdef FAST_LOADER
		if (!loader_fast_fl)
#endif
		    loader_send(znak);


		suma = (int)znak;
		bytu_inline = (int)znak - 5;

		adresa = 0;
		for(i=0; i < 4; i++) {
			znak = loader_recv();
#ifdef FAST_LOADER
			if (!loader_fast_fl)
#endif
			    loader_send(znak);

			suma += (int)znak;
			adresa = adresa << 8;
			adresa += ((unsigned long)znak) & 0xFF;
		}

/*		LOADER_PRINTF("\nAdr=%p", (char *)adresa); */
		
		// Konec Downloadu, posle OK
		if(adresa == 0xffffffff) {
			//KONEC, zapis!!
			if(buff_cnt == 0) {
				buff_cnt = -1;
			
				if (In_flash == 1)
				    r = ((tFwrite_INT)RAM_buf)(adresa_buff, (unsigned long *)FL_buff, 1);
				else if (In_flash == 3) {
				    r = 0;
				    ide_write(0, (adresa_buff - RAM_EXT_START)/FLASH_CF_BLK_SIZE + FLASH_CF_BOOT_SIZE,
				     1, (ulong *)FL_buff);
				} else
				    r = Fwrite_INT(adresa_buff, (unsigned long *)FL_buff, 0);

				if(r < 0)return(-3);
			} /*endif*/

			znak = loader_recv();
#ifdef FAST_LOADER
			/* echo sumy posila pouze za poslednim prijatym radkem */
			if (!loader_IsChar() || (!loader_fast_fl))
#endif
			loader_send(znak);	// kontrolni soucet konce

			loader_send('O');
			loader_send('K');
			return(0);
		} /*endif*/

		/* Pokud data, tak je nactu */
		if (bytu_inline) {
#ifdef FAST_LOADER
			int magic = ((int)serial[13] & 0xff) - 28;

			for(i=0; i < bytu_inline; i++, adresa++, magic++) {
				if (loader_fast_fl) {
				    znak = loader_recv() ^ magic;
//				    LOADER_PRINTF("[%ld]=%02lX,", (long)i, (long)znak);
				} else {
				    znak = loader_recv();
				    loader_send(znak);
				}
#else
			for(i=0; i < bytu_inline; i++, adresa++) {
				znak = loader_recv();
				loader_send(znak);
#endif
				suma += (unsigned int)znak;

				//Interni Flash
				if((adresa >= FLASH_START) && (adresa <= FLASH_INT_TOP) && (In_flash != 2)) {
					// 1-Ulozeni platneho bloku
					//vyjel jsem z bloku, zapis!!
					if(!buff_cnt && (adresa_buff != (adresa & ~(FLASH_INT_BLK_SIZE-1)))) {
						buff_cnt = -1;
						
						if(In_flash == 1)
							r = ((tFwrite_INT)RAM_buf)(adresa_buff, (unsigned long *)FL_buff, 1);
						else {
							r = Fwrite_INT(adresa_buff, (unsigned long *)FL_buff, 0);
						}
							
						if(r < 0) return(-3);
					}

					// 2 - Nastaveni noveho bloku
					if(buff_cnt == -1) {
						// Adresa bloku
						adresa_buff = adresa & ~(FLASH_INT_BLK_SIZE-1);
						buff_cnt = 0;
					
					 	// naplnim si pole tim, co tam bylo, kdybych zapisoval jen cast
						for(x=0;x<(FLASH_INT_BLK_SIZE/sizeof(uint32));x++)
							((uint32*)FL_buff)[x] = ((uint32*)adresa_buff)[x];
					}
 
					// 3 - ulozim znak na pozici buferu
					//ulozim byte??
					if(adresa_buff == (adresa & ~(FLASH_INT_BLK_SIZE-1)))
						FL_buff[adresa & (FLASH_INT_BLK_SIZE-1)] = znak;
				} //konec intern

/* Externi SDRAM - nyni uklada vse co je urceno pro SDRAM do Compact Flash */

/*				if((adresa >= RAM_EXT_START) && (adresa <= RAM_EXT_TOP) && (In_flash == 2)) { */
				if((adresa >= RAM_EXT_START) && (adresa <= RAM_EXT_TOP)) {
#if 0
					RAMwrite_EXT(adresa, znak);
#else
					int kk;

					/* nastavi flag, ze uklada do Compact Flash */
					In_flash = 3;

					// 1-Ulozeni platneho bloku
					//vyjel jsem z bloku, zapis!!
					if(!buff_cnt && (adresa_buff != (adresa & ~(FLASH_CF_BLK_SIZE-1)))) {
						buff_cnt = -1;
						
						/* r = Fwrite_CF(adresa_buff, (unsigned long *)FL_buff, 0); */
						ide_write(0, (adresa_buff - RAM_EXT_START)/FLASH_CF_BLK_SIZE + FLASH_CF_BOOT_SIZE,
						 1, (ulong *)FL_buff);

						/* provede i verifikaci */
						ide_read(0, (adresa_buff - RAM_EXT_START)/FLASH_CF_BLK_SIZE + FLASH_CF_BOOT_SIZE,
						 1, (ulong *)FL_buff2);

						for(r = 0, kk = 0; !r && (kk < FLASH_CF_BLK_SIZE/sizeof(ulong)); kk++) {
							if (((ulong *)FL_buff)[kk] != ((ulong *)FL_buff2)[kk])
								r = -3;
						} /*for*/

						if(r < 0)
							return(-3);
					} /*endif*/

					// 2 - Nastaveni noveho bloku
					if(buff_cnt == -1) {
						// Adresa bloku
						adresa_buff = adresa & ~(FLASH_CF_BLK_SIZE-1);
						buff_cnt = 0;
						
					 	// naplnim si pole tim, co tam bylo, kdybych zapisoval jen cast
						/* Fread_CF - cteni sektoru z Compact Flash */
/*
						for(x = 0; x < (FLASH_CF_BLK_SIZE/sizeof(uint32)); x++)
							((uint32*)FL_buff)[x] = ((uint32*)adresa_buff)[x];
*/
						ide_read(0, (adresa_buff - RAM_EXT_START)/FLASH_CF_BLK_SIZE + FLASH_CF_BOOT_SIZE,
						 1, (ulong *)FL_buff);
					} /*endif*/
 
					// 3 - ulozim znak na pozici buferu
					//ulozim byte??
					if(adresa_buff == (adresa & ~(FLASH_CF_BLK_SIZE-1)))
						FL_buff[adresa & (FLASH_CF_BLK_SIZE-1)] = znak;
#endif
				} //konec Compact Flash

/* externi Flash  */
#if 0
				// Externi Flash
				// Ted pisu do vnejsi tak vnitrni ulozim, pokud je co
				if((adresa >= FLASH_EXT_START) && (adresa <= FLASH_TOP) && (In_flash != 2)) {
					//KONEC, zapis!!
					if(!buff_cnt) {
						buff_cnt = -1;
						
         				if(In_flash == 1)
							r = ((tFwrite_INT)RAM_buf)(adresa_buff, (unsigned long *)FL_buff, 1);
						else
							r = Fwrite_INT(adresa_buff, (unsigned long *)FL_buff, 0);
						
						if(r < 0)return(-3);
					}
				
	        		// Zapis ven
					Fwrite_EXT( adresa, znak );
				} /*endif*/
#endif /*0*/
			} /*for*/
		} /*endif*/
		
		// Kontrola souctu
//		LOADER_PRINTF("<<");
		znak = loader_recv();
//		LOADER_PRINTF(">>");

#ifdef FAST_LOADER
		/* echo sumy posila pouze za poslednim prijatym radkem */
		if (!loader_IsChar() || (!loader_fast_fl)) {
#endif
		    loader_send(znak);
#ifdef FAST_LOADER
//		    LOADER_PRINTF("**");
		} else {
//		    LOADER_PRINTF("++");
		}
#endif

		suma += (unsigned int)znak;
		suma = suma & 0xff;
		
		if(suma != 0xff)
			return(-2);

	} while(1);// Konec loader smycky
} // Konec Loaderu

/*******************************************************************/
/* FCE PRO VYSLANI A PRIJEM UDP PAKETU PO ETHERNETU                */
/******************************************************************/
/*****/
/* obsluzna rutina pri prijmu meho UDP paketu */
void
loader_udp_handler (NIF *nif, NBUF *pNbuf, int hdr_offset)
{
	(void)nif;
	int	i;

	/* Accessing shared memory here, disable interrupts */
	asm_set_ipl(6);

	/* nacteme UDP paket do bufferu */
//	LOADER_PRINTF("\nLH: ln=%ld,char=", (long)pNbuf->length);

	for(i = 0; i < pNbuf->length; i++) {
	    recv_char[recv_char_tail] = pNbuf->data[hdr_offset + i];
//	    LOADER_PRINTF("%02lX,", (long)recv_char[recv_char_tail]);

	    if (++recv_char_tail >= cLOADER_RXBUF_LN)
		recv_char_tail = 0;
	}

	recv_char_ln += pNbuf->length;
	recv_char_fl = TRUE;
	
	/* Re-enable interrupts now */
//	asm_set_ipl(0);
} /* end of loader_udp_handler() */

/*****/
/* nacte paket dat zadane delky
 * pokud neni pozadovany pocet znaku
 * k dispozici, vyprsi timeout
 * v promenne ln vraci pocet nactenych znaku
 * vraci: LOADER_RETOK
 *        LOADER_ERR_TOUT   - vyprsel timeout pri cekani na znak
 *        LOADER_ERR_LENGTH - v bufferu zustaly neprectene znaky
*/
static int
loader_recv_paket(
  uint8		*buff,
  uint32	*ln)
{
    uint32	i;
    uint32	tout;

    for(i = 0; i < *ln; i++) {
	/* ceka na prijaty znak, hlida timeout */
	tout = 0;
	while(!loader_IsChar() && (tout < cLOADER_RECV_TOUT)) {
	    udelay(1000);
	    tout++;
	} /*while*/

	/* pokud neprisel od PC dalsi znak, konec */
	/* jinak ho uloz do bufferu */
	if (loader_IsChar())
	    buff[i] = loader_recv();
	else
	    break;
    } /*for*/

    /* paket ma jinou nez pozadovanou delku */
    if (i != *ln) {
	*ln = i;

	return(LOADER_ERR_TOUT);
    } else if (loader_IsChar())
	return(LOADER_ERR_LENGTH);
    else
	return(LOADER_RETOK);
} /* end of loader_recv_paket() */

/*****/
/* nacte paket dat zadane delky
 * [1] pokud neni pozadovany pocet znaku
 *     k dispozici, vyprsi timeout
 *     v promenne ln vraci pocet nactenych znaku
 * [2] provede kontrolu CRC, pokud je v poradku
 *     a ack == TRUE, posle CRC zpet - 4 byty
 * [3] pokud doslo k chybe, posle zpet kod chyby - 2 byty
 *
 * vraci: LOADER_RETOK
 *        LOADER_ERR_TOUT   - vyprsel timeout pri cekani na znak
 *        LOADER_ERR_LENGTH - v bufferu zustaly neprectene znaky
 *        LOADER_ERR_CRC    - chyba kontrolniho souctu
*/
static int
loader_recvchk_paket(
  uint8		*buff,
  uint32	*ln,
  int		ack)		/* ==TRUE, posle potvrzeni */
{
    uint32	l = *ln + sizeof(uint32);
    uint32	ret32, crc;
    uint16	ret16;

    if (loader_recv_paket(buff, &l) != LOADER_RETOK) {
	LOADER_PRINTF("\nrecv paket: length error");

	*ln = l;

	ret16 = cLOADER_PCERR_LENGTH;

	loader_send_paket((uint8 *)&ret16, sizeof(uint16));

	return(LOADER_ERR_LENGTH);
    } /*endif*/

    /* zkontroluje CRC */
    if ((crc = crc32(0, buff, *ln)) != *((uint32 *)(buff + *ln))) {
	LOADER_PRINTF("\nrecv paket: CRC error");

        ret16 = cLOADER_PCERR_CRC;

	loader_send_paket((uint8 *)&ret16, sizeof(uint16));

	return(LOADER_ERR_CRC);
    } /*endif*/

    /* CRC i delka odpovidaji, posle zpet potvrzeni prijmu
     * nyni jako echo CRC
     */
    if (ack) {
	/* vse v poradku, posli CRC zpet */
	ret32 = crc;
	loader_send_paket((uint8 *)&ret32, sizeof(uint32));
    } /*endif*/

    return(LOADER_RETOK);
} /* end of loader_recvchk_paket() */

/*****/
/* vysle paket dat.
 * pokud vysila na ser. linku, posle 
 * paket po znacich,
 * pokud komunikuje pres UDP,
 * rozseka data na UDP pakety
*/
static int
loader_send_paket(
  uint8		*buff,
  uint32	ln)
{
	int	i;

    if (loader_send == loader_udp_send) {
	NBUF	*pNbuf;

	/* pozada o vysilaci buffer */
	pNbuf = fec_nif.tx_alloc();

	/* Make sure I grabbed a valid Tx buffer */
	if (pNbuf == NULL) {
	    printf("\nloader_send() couldn't allocate Tx buffer");
	    return 0;
	}

	/* store char to buffer */
	for (i = 0; (i < ln) && (i < 1000); i++)
	    pNbuf->data[LOADER_HDR_OFFSET + i] = buff[i];

	pNbuf->length = ln;

	return udp_send(&fec_nif, server,
       	 LOADER_UDP_PORT,		/* source port */
	 LOADER_UDP_PORT,		/* destination port */
	 pNbuf);
    } else {
	/* store char to buffer */
	for (i = 0; i < ln; i++)
	    uart_send(buff[i]);

	return(LOADER_RETOK);
    } /*endif*/
} /* end of loader_send_paket() */

/*****/
/* vysle znak pomoci UDP paketu */
int
loader_udp_send(char znak)
{
	NBUF *pNbuf;

	/* pozada o vysilaci buffer */
	pNbuf = fec_nif.tx_alloc();

	/* Make sure I grabbed a valid Tx buffer */
	if (pNbuf == NULL)
	{
	    printf("\nloader_send() couldn't allocate Tx buffer");
	    return 0;
	}

	/* store char to buffer */
	pNbuf->data[LOADER_HDR_OFFSET] = znak;

	pNbuf->length = 1;

	return udp_send(&fec_nif, server,
					LOADER_UDP_PORT,		/* source port */
					LOADER_UDP_PORT,		/* destination port */
					pNbuf);
} /* end of loader_udp_send() */

/* prijme znak pomoci UDP paketu */
char
loader_udp_recv(void)
{
    char znak;

    while(!loader_udp_IsChar());

    /* smaze flag "vyzvednut posledni znak z bufferu" */
    if (--recv_char_ln == 0)
	recv_char_fl = FALSE;

    znak = recv_char[recv_char_head++];

    if (recv_char_head >= cLOADER_RXBUF_LN)
	recv_char_head = 0;

//    LOADER_PRINTF("R%02lX,", (long)znak);
    return(znak);
} /* end of loader_udp_recv() */

/*****/
int
loader_udp_IsChar(void)
{
    /* obslouzi prijem eth paketu */
    if (MCF5282_FEC_EIR & (MCF5282_FEC_EIR_RXB | MCF5282_FEC_EIR_RXF))
    {
	// printf("\nFEC_EIR=0x%08lX", (long)MCF5282_FEC_EIR);
	/* Clear interrupt from EIR register immediately */
	MCF5282_FEC_EIR = (MCF5282_FEC_EIR_RXB | MCF5282_FEC_EIR_RXF);

	fec_nif.receive(&fec_nif);
    }

    return(recv_char_ln != 0);
} /* end of loader_udp_IsChar() */

/*******************************************************************/
#if 0
// Funkce pro odeslani a prijmuti znaku 
/* FCE pro komunikaci po ser. lince */
int send(char znak) 
{
    return(uart_send(znak));
}
#endif
#if 1
char
recv(void)
{
  return(uart_recv());
#if 0
    do{}while(!IsChar());   /* Loop if rx buffer empty */

	/* Get Received data */
    return(MCF5282_UART0_URB);
#endif
}

// Vraci 1, pokud Char ready
int
IsChar(void)
{
    return(uart_IsChar());
}
#endif

char
uart_recv(void)
{
    do{}while(!uart_IsChar());   /* Loop if rx buffer empty */

    /* Get Received data */
    return(MCF5282_UART0_URB);
}

int uart_send(char znak) 
{
    /* Loop until transmitter holding register not empty */
    do{}while(!(MCF5282_UART0_USR & MCF5282_UART_USR_TXRDY));

	MCF5282_UART0_UTB = znak; /* Send data */
	return(0);
}

// Vraci 1, pokud Char ready
int
uart_IsChar(void)
{
    /* obslouzi prijem eth paketu */
    if (MCF5282_FEC_EIR & (MCF5282_FEC_EIR_RXB | MCF5282_FEC_EIR_RXF))
    {
	// printf("\nFEC_EIR=0x%08lX", (long)MCF5282_FEC_EIR);
	/* Clear interrupt from EIR register immediately */
	MCF5282_FEC_EIR = (MCF5282_FEC_EIR_RXB | MCF5282_FEC_EIR_RXF);

	fec_nif.receive(&fec_nif);
    }

    return((MCF5282_UART0_USR & MCF5282_UART_USR_RXRDY) != 0);
}

/*<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
/*--------------------------------------------------------------*/
/* Zakladni fce pro operace s pametmi Interni FLASH SuperFlash   */
// Maze Interni Flash, 
int
FeraseA_INT(unsigned int In_flash)
{
	uint32	count,xx;
	uint8	*__FLASH_ERASE_START;

	count = 0;
	
	if (!(MCF5282_CFM_CFMCLKD & MCF5282_CFM_CFMCLKD_DIVLD))
		return(LOADER_RET_ERR_ERASE);

	//******************************************************
	// Automaticka detekce umisteni, pokud sedi ve FLASH, tak svuj sektor nemaze!
	// (pouzije prikaz "Page Erase"). Jinak provede prikaz "Mass Erase" na oba bloky
	if(In_flash == 1) {
/*		__FLASH_ERASE_START = (uint8 *)&__FLASH[FLASH_LOADER_TOP + 1]; */
/*		__FLASH_ERASE_START = (uint8 *)0x44000000; */
		__FLASH_ERASE_START = (uint8 *)&__IPSBAR[0x04000000+FLASH_LOADER_SIZE];

		while((uint32)__FLASH_ERASE_START < (uint32)(__IPSBAR+0x04000000+FLASH_SIZE)) {
			/* preskoci blok, v nemz jsou ulozeny info o S/N */
			if (((uint32)__FLASH_ERASE_START - (uint32)(__IPSBAR+0x04000000) + FLASH_ADDRESS)
			 == ((uint32)psafe_info & ~(FLASH_INT_PAGE_ERASE_BLOCK_SIZE - 1))) {
			    __FLASH_ERASE_START += FLASH_INT_PAGE_ERASE_BLOCK_SIZE;

			    continue;
			} /*endif*/

			/* ceka na dokonceni predchozi operace Program,Verify,Erase */
			xx = 0;
			while (!(MCF5282_CFM_CFMUSTAT &
			 (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR |
			 MCF5282_CFM_CFMUSTAT_CCIF)) && (xx++ < 5000000));
			//IFLASH_WAIT_FOR_EPV_RDY;
			
			/* check for errors */
			if ((MCF5282_CFM_CFMUSTAT & (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR)) |
				(xx >= 5000000)) {
				MCF5282_CFM_CFMUSTAT |= (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR);
				
				return(LOADER_RET_ERR_ERASE);
			} /*endif*/
			
			/* Check, if page is already erased by means of "Page Erase Verify command" */
			/* clear BLANK bit in status register */
			MCF5282_CFM_CFMUSTAT |= MCF5282_CFM_CFMUSTAT_BLANK;

			/* write dummy word for set page erase address */
			*((uint32 *)__FLASH_ERASE_START) = 0x55aa55aa;
		
			/* set "Page Erase Verify" Command */
			IFLASH_SET_CMD(IFLASH_CMD_VERIFY_PAGE);
		
			/* start page erase verification */
			IFLASH_START_CMD;
		
			/* ceka na dokonceni predchozi operace Program,Verify,Erase */
			xx = 0;
			while (!(MCF5282_CFM_CFMUSTAT &
			 (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR |
			 MCF5282_CFM_CFMUSTAT_CCIF)) && (xx++ < 5000000));

			/* check for errors */
			if ((MCF5282_CFM_CFMUSTAT & (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR)) |
				(xx >= 5000000)) {
				MCF5282_CFM_CFMUSTAT |= (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR);
				
				return(LOADER_RET_ERR_ERASE);
			} /*endif*/

			/* Erase page only if not already erased */
			if (!(MCF5282_CFM_CFMUSTAT & MCF5282_CFM_CFMUSTAT_BLANK)) {
				/* write dummy word for set page erase address */
				*((uint32 *)__FLASH_ERASE_START) = 0x55aa55aa;
		
				/* set Page Erase Command */
				IFLASH_SET_CMD(IFLASH_CMD_ERASE_PAGE);
		
				/* start page erase */
				IFLASH_START_CMD;
			} /*endif*/
		
			__FLASH_ERASE_START += FLASH_INT_PAGE_ERASE_BLOCK_SIZE;
		} /*while*/
		
		/* ceka na dokonceni operace "Page Erase" */
		xx = 0;
		while (!(MCF5282_CFM_CFMUSTAT &
		 (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR |
		 MCF5282_CFM_CFMUSTAT_CCIF)));
		
		/* check for errors */
		if (MCF5282_CFM_CFMUSTAT & (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR)) {
			MCF5282_CFM_CFMUSTAT |= (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR);
			
			return(LOADER_RET_ERR_ERASE);
		} /*endif*/
			
		//IFLASH_WAIT_FOR_EPV_RDY;
	} else {
		__FLASH_ERASE_START = (uint8 *)__FLASH;
		__FLASH_ERASE_START = (uint8 *)0x44000000;
		
		/* Loader in RAM, so erase all internal Flash by means of 
		 * "Mass Erase" command
		*/
		while((uint32)__FLASH_ERASE_START < (uint32)(__IPSBAR+0x04000000+FLASH_SIZE)) {
			/* ceka na dokonceni predchozi operace Program,Verify,Erase */
			while (!(MCF5282_CFM_CFMUSTAT &
			 (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR |
			 MCF5282_CFM_CFMUSTAT_CCIF)));
			//IFLASH_WAIT_FOR_EPV_RDY;
			
			/* check for errors */
			if (MCF5282_CFM_CFMUSTAT & (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR)) {
				/* clear PVIOL and ACCERR bits in CFMUSTAT */
				MCF5282_CFM_CFMUSTAT |= (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR);
				
				return(LOADER_RET_ERR_ERASE);
			} /*endif*/
			
			/* write dummy word for set page erase address */
			*((uint32 *)__FLASH_ERASE_START) = 0x55aa55aa;
		
			/* set "Mass Erase" Command */
			IFLASH_SET_CMD(IFLASH_CMD_ERASE_MASS);
		
			/* start Mass Erase */
			IFLASH_START_CMD;
		
			__FLASH_ERASE_START += FLASH_INT_MASS_ERASE_BLOCK_SIZE;
		} /*while*/
		
		/* ceka na dokonceni operace "Mass Erase" */
		while (!(MCF5282_CFM_CFMUSTAT &
		 (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR |
		 MCF5282_CFM_CFMUSTAT_CCIF)));
		
		/* check for errors */
		if (MCF5282_CFM_CFMUSTAT & (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR)) {
			/* clear PVIOL and ACCERR bits in CFMUSTAT */
			MCF5282_CFM_CFMUSTAT |= (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR);
			
			return(LOADER_RET_ERR_ERASE);
		} /*endif*/
			
		//IFLASH_WAIT_FOR_EPV_RDY;
	} /*endif*/

	return(LOADER_RETOK);
} /* end of FeraseA_INT() */

/*****/
//Zapisuje do obou bloku A,B 64bytove bloky, adresa musi byt zarovnana na 0x3F
int
Fwrite_INT(long adr, unsigned long *buffer_64byte, unsigned int In_flash)
{
	vuint32		*BlockA;
	uint32		xx;
	int			i;

	// Base Flash pro programovani
	BlockA = (volatile unsigned long *)&__IPSBAR[0x04000000L];

	for(i = 0; i <  FLASH_INT_BLK_SIZE/sizeof(uint32); i++, adr += sizeof(uint32)) {
		// Nejdrive si zjistim, do ktereho segmnetu chci zapisovat, spolecne pro A,B
		if(adr > FLASH_INT_TOP)
			return(-1);

		/* nic pro mne, neprepisu prece loader */
		if((In_flash == 1) && (adr < FLASH_LOADER_SIZE))
			return(-1);

		/* ceka na dokonceni predchozi operace Program,Verify,Erase */
		xx = 0;
		while (!(MCF5282_CFM_CFMUSTAT &
		 (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR |
		 MCF5282_CFM_CFMUSTAT_CCIF)) && (xx++ < 5000000));
			
		/* check for errors */
		if ((MCF5282_CFM_CFMUSTAT & (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR)) |
			(xx >= 5000000)) {
			MCF5282_CFM_CFMUSTAT |= (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR);
				
			return(LOADER_RET_ERR_ERASE);
		} /*endif*/
			
		/* write data and address to buffer */
		BlockA[(adr-FLASH_START)/sizeof(vuint32)] = buffer_64byte[i];
		
		/* set "Program" Command */
		IFLASH_SET_CMD(IFLASH_CMD_PROGRAM);

		/* start programming */
		IFLASH_START_CMD;
		
		/* ceka na dokonceni operace Program */
		xx = 0;
		while (!(MCF5282_CFM_CFMUSTAT &
		 (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR |
		 MCF5282_CFM_CFMUSTAT_CCIF)) && (xx++ < 5000000));
			
		/* check for errors */
		if ((MCF5282_CFM_CFMUSTAT & (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR)) |
			(xx >= 5000000)) {
			MCF5282_CFM_CFMUSTAT |= (MCF5282_CFM_CFMUSTAT_PVIOL | MCF5282_CFM_CFMUSTAT_ACCERR);
				
			return(LOADER_RET_ERR_ERASE);
		} /*endif*/
	} /*for*/

	// Verify
#if 0
	xx = 0;
	do{}while((BlockA[adr/sizeof(vuint32)] != buffer_64byte[0]) && (xx++ < 5000000));
#endif /*0*/

/*	if (xx >= 5000000) {
		return(LOADER_RET_ERR_ERASE);
	} else { */
		return(LOADER_RETOK);
/*	}*/
} /* end of Fwrite_INT() */

/**************************************************************/
/** Verifikuje data ulozena do Interni Flash                  */
int Fverify_INT(long adr, unsigned long *buffer_64byte)
{
volatile unsigned long *BlockA,*BlockB;
unsigned long x;

BlockA = (volatile unsigned long *)0x00000L; // Base blocku A
BlockB = (volatile unsigned long *)0x40000L; // Base blocku B


// Copy data in to 64buffer
for(x=0;x<(64/4);x++) BlockA[(adr/4)+x] = buffer_64byte[x];

// Verify 64bytes
x = 0;
do{

 if( BlockA[(adr/4)+x] != buffer_64byte[x] )
  return(-1);

 x++;
}while( x < (64/4) );

return(0); // Probehlo vse OK
}


/****************************************************************/
/****************************************************************/
/* Zakladni fce pro operace s pametmi Externi FLASH AMD29LV017  */
// Jeste nejsou upravene

//#define ADR_FLS(a,w) (a)*SLOVO_EXT_LEN + (w) 
//#define ADR_FLS(a,w) ( ((a & 1) * 0x400000 ) + ((a>>1)*SLOVO_EXT_LEN) + (w) )
#define ADR_FLS(a,w)  ((((a) & 1)?(1 << 21):0) | (((a) & ~1) + (w)))

int Ferase_EXT( int word ) // zadava se cislo bajtu
{
	volatile unsigned char *Pamet;
	Pamet = (volatile unsigned char *)FLASH_EXT_START;

	// Mazaci sekvence	
	Pamet[ ADR_FLS(0x5555,word) ] = 0xAA;
	Pamet[ ADR_FLS(0x2AAA,word) ] = 0x55;	
	Pamet[ ADR_FLS(0x5555,word) ] = 0x80;  //Erase	
	Pamet[ ADR_FLS(0x5555,word) ] = 0xAA;	
	Pamet[ ADR_FLS(0x2AAA,word) ] = 0x55;	
	Pamet[ ADR_FLS(0x5555,word) ] = 0x10;	

	return(0);
}

int Ferase_EXT_WF( int word ) // zadava se cislo bajtu
{
	unsigned char x,xx;

	volatile unsigned char *Pamet;
	Pamet = (volatile unsigned char *)FLASH_EXT_START;

	// Toogle Algorithm
	do{
		x  = Pamet[ ADR_FLS(0x5555,word) ];	
		xx = Pamet[ ADR_FLS(0x5555,word) ];	
	}while((x & 0x40) != (xx & 0x40));
	return(0);
}

int Fwrite_EXT(long adr, unsigned char cislo)
{
	unsigned char x,xx;
	int word;

	volatile unsigned char *Pamet;
	Pamet = (volatile unsigned char *)FLASH_EXT_START;
	adr -= FLASH_EXT_START;

	word = adr % SLOVO_EXT_LEN;

	Pamet[ ADR_FLS(0x5555,word) ]= 0xAA;	
	Pamet[ ADR_FLS(0x2AAA,word) ]= 0x55;	
	Pamet[ ADR_FLS(0x5555,word) ]= 0xA0;  //Write	
	
	Pamet[ adr ]= cislo;	

	// Toogle Algorithm
	do {
		x =  Pamet[ adr ];	
		xx = Pamet[ adr ];	
	} while((x & 0x40) != (xx & 0x40));

	return(0);
}

/**********************************************************/
/**********************************************************/
/*  Fce pro zapis byte do externi RAM                     */
int RAMwrite_EXT(long adresa,unsigned char znak )
{
	volatile unsigned char *Pamet;
	Pamet = (volatile unsigned char *)RAM_EXT_START;

	Pamet[adresa - RAM_EXT_START] = znak;	
	
	return(0);
}

/******************************************************/
/*    Utilities                                       */

/* Kopiruje blok dat */
int copy(char *s,char *d, unsigned int ln)
{char *ss,*dd;
	unsigned int i;
ss = s; dd = d;
	
	for(i=0;i<ln;i++)*dd++ = *ss++;
return(0);	
}


// Vypise string
void Vyslitext(char *t)
{
int i;

i=0;
do{ 
    uart_send(t[i]); 
  }while(t[i++]!=0);

}

// Vypise LONG na Serial
void VysliLong (long b)
{
 int a;
 uart_send('$');
 a=(b & 0xf0000000L)>>28;
 a+='0';
 if(a>'9')a+=7;
 uart_send(a);
 a=(b & 0xf000000L)>>24;
 a+='0';
 if(a>'9')a+=7;
 uart_send(a);
 a=(b & 0xf00000L)>>20;
 a+='0';
 if(a>'9')a+=7;
 uart_send(a);
 a=(b & 0xf0000L)>>16;
 a+='0';
 if(a>'9')a+=7;
 uart_send(a);
 a=(b & 0xf000L)>>12;
 a+='0';
 if(a>'9')a+=7;
 uart_send(a);
 a=(b & 0xf00L)>>8;
 a+='0';
 if(a>'9')a+=7;
 uart_send(a);
 a=(b & 0xf0L)>>4;
 a+='0';
 if(a>'9')a+=7;
 uart_send(a);
 a=(b & 0xfL);
 a+='0';
 if(a>'9')a+=7;
 uart_send(a);
}
